# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that transform the instance in some
way"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from ase import Atoms
from ase.quaternions import Quaternion
from soprano.utils import minimum_periodic
from soprano.properties import AtomsProperty
from soprano.selection import AtomSelection


def _transform_sel_check(extrfunc):

    def decorated_extrfunc(s, selection, **kwargs):

        # Perform basic checks on selection
        if selection is None:
            selection = AtomSelection.all(s)
        elif not selection.validate(s):
            raise ValueError('Selection passed to transform does not apply to'
                             ' system.')

        return extrfunc(s, selection, **kwargs)

    return decorated_extrfunc


class Translate(AtomsProperty):

    """
    Translate

    Returns an Atoms object with some or all the atoms translated by a given
    vector. Absolute or scaled coordinates may be used.

    | Parameters:
    |   selection (AtomSelection): selection object defining which atoms to
    |                              act on. By default, all of them.
    |   vector ([float]*3): vector by which to translate the atoms.
    |   scaled (bool): if True, treat the input vector as expressed in scaled,
    |                  not absolute, coordinates.

    | Returns:
    |   translated (ase.Atoms): Atoms object with the translation performed.

    """

    default_name = "translated"
    default_params = {
        'selection': None,
        'vector': [0, 0, 0],
        'scaled': False
    }

    @staticmethod
    @_transform_sel_check
    def extract(s, selection, vector, scaled):

        vector = np.array(vector)
        if vector.shape != (3,):
            raise ValueError('Invalid vector passed to Translate.')

        sT = s.copy()

        if not scaled:
            pos = sT.get_positions()
        else:
            pos = sT.get_scaled_positions()

        pos[selection.indices] += vector

        if not scaled:
            sT.set_positions(pos)
        else:
            sT.set_scaled_positions(pos)

        return sT


class Rotate(AtomsProperty):

    """
    Rotate

    Returns an Atoms object with some or all the atoms rotated by a given
    quaternion and with a given center. Absolute or scaled coordinates may be
    used.

    | Parameters:
    |   selection (AtomSelection): selection object defining which atoms to
    |                              act on. By default, all of them.
    |   center ([float]*3): center around which the rotation should take
    |                       place. By default the origin of the axes.
    |   quaternion (ase.quaternions.Quaternion): quaternion expressing the
    |                                            rotation that should be
    |                                            applied.
    |   scaled (bool): if True, treat the input vector as expressed in scaled,
    |                  not absolute, coordinates.
    |   periodic (bool): if True, always rotate each atom around the minimum
    |                    periodic image of the rotation center.

    | Returns:
    |   rotated (ase.Atoms): Atoms object with the rotation performed.

    """

    default_name = "rotated"
    default_params = {
        'selection': None,
        'center': [0, 0, 0],
        'quaternion': None,
        'scaled': False,
        'periodic': False,
    }

    @staticmethod
    @_transform_sel_check
    def extract(s, selection, center, quaternion, scaled, periodic):

        center = np.array(center)
        if center.shape != (3,):
            raise ValueError('Invalid center passed to Rotate.')

        if quaternion is None:
            quaternion = Quaternion()

        sT = s.copy()

        if not scaled:
            pos = sT.get_positions()
        else:
            pos = sT.get_scaled_positions()

        pos -= center
        if periodic:
            ppos, _ = minimum_periodic(pos[selection.indices], s.get_cell())
            pos[selection.indices] = ppos
        pos[selection.indices] = quaternion \
            .rotate(pos[selection.indices].T).T
        pos += center

        if not scaled:
            sT.set_positions(pos)
        else:
            sT.set_scaled_positions(pos)

        return sT


class Mirror(AtomsProperty):

    """
    Mirror

    Returns an Atoms object with some or all the atoms reflected with either a
    given center or a given plane. Absolute or scaled coordinates may be used.

    | Parameters:
    |   selection (AtomSelection): selection object defining which atoms to
    |                              act on. By default, all of them.
    |   center ([float]*3): center around which the reflection should take
    |                       place. By default the origin of the axes. Can't be
    |                       present at the same time as plane.
    |   plane ([float]*4): plane with respect to which the reflection should
    |                      take place, in the form [a, b, c, d] parameters of
    |                      the plane equation.
    |                      By default is not used. Can't be present at the
    |                      same time as center.
    |   scaled (bool): if True, treat the input vector as expressed in scaled,
    |                  not absolute, coordinates.

    | Returns:
    |   reflected (ase.Atoms): Atoms object with the reflection performed.

    """

    default_name = "reflected"
    default_params = {
        'selection': None,
        'center': None,
        'plane': None,
        'scaled': False
    }

    @staticmethod
    @_transform_sel_check
    def extract(s, selection, center, plane, scaled):

        if plane is not None and center is not None:
            raise ValueError('Can\'t pass both center and plane to Mirror')

        if center is not None:
            center = np.array(center)
            if center.shape != (3,):
                raise ValueError('Invalid center passed to Mirror.')
        elif plane is not None:
            plane = np.array(plane)
            if plane.shape != (4,):
                raise ValueError('Invalid plane passed to Mirror.')
        else:
            center = np.zeros(3)

        sT = s.copy()

        if not scaled:
            pos = sT.get_positions()
        else:
            pos = sT.get_scaled_positions()

        if center is not None:

            pos -= center
            pos[selection.indices] *= -1
            pos += center

        else:

            # Find the components of the position vectors normal to the plane
            nu = plane[:3]/np.linalg.norm(plane[:3])
            norm_pos = (np.dot(pos[selection.indices],
                               plane[:3])+plane[3])[:, None]*nu
            pos[selection.indices] = pos[selection.indices] - 2*norm_pos

        if not scaled:
            sT.set_positions(pos)
        else:
            sT.set_scaled_positions(pos)

        return sT



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
utils.py

Contains package-wide useful routines that don't fall under any specific
category. Many of these handle common operations involving periodicity,
conversions between different representations etc.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
from contextlib import contextmanager
import inspect
import warnings
import numpy as np
from ase import Atoms
from scipy.special import factorial
from ase.quaternions import Quaternion
from itertools import product as iter_product

from soprano.optional import (requireNetworkX, requireScikitLearn,
                              requireSpglib)


def seedname(path):
    """Get the filename (with no extension) from a full path"""
    return os.path.splitext(os.path.basename(path))[0]


def replace_folder(path, new_folder):
    """Replace the folder of the given path with a new one"""
    return os.path.join(new_folder, os.path.basename(path))


def progbar(i, i_max, bar_len=20, spinner=True, spin_rate=3.0):
    """A textual progress bar for the command line

    | Args:
    |   i (int): current progress index
    |   max_i (int): final progress index
    |   bar_len (Optional[int]): length in characters of the bar (no brackets)
    |   spinner (Optional[bool]): show a spinner at the end
    |   spin_rate (Optional[float]): spinner rotation speed (turns per full
    |                                progress)

    | Returns:
    |   bar (str): a progress bar formatted as requested

    """
    block = {True: '\u2588', False: ' '}
    spin = '|\\-/'
    perc = i/float(i_max)*bar_len
    bar = '[{0}]'.format(''.join([block[i < perc] for i in range(bar_len)]))
    if spinner:
        bar += ' {0}'.format(spin[int(perc*spin_rate) % len(spin)])

    return bar


@contextmanager
def silence_stdio(silence_stdout=True, silence_stderr=True):
    """Useful stdout/err silencer"""
    dummy_out = open(os.devnull, "w")
    if silence_stdout:
        old_stdout, sys.stdout = sys.stdout, dummy_out
    if silence_stderr:
        old_stderr, sys.stderr = sys.stderr, dummy_out
    try:
        yield dummy_out
    finally:
        if silence_stdout:
            sys.stdout = old_stdout
        if silence_stderr:
            sys.stderr = old_stderr


def customize_warnings():

    def customwarning(msg, category, filename, lineno, line=None):
        outmsg = ('\033[93m \033[1m WARNING: \033[0m {0} '
                  '({1}, line: {2})\n').format(msg, filename, lineno)
        return outmsg

    warnings.formatwarning = customwarning

# Apply it right away (for stuff in this module)
customize_warnings()

def abc2cart(abc):
    """Transforms an axes and angles representation of lattice parameters
       into a Cartesian one

    """

    abc = np.array(abc, copy=False)

    if abc.shape != (2, 3):
        raise ValueError("Invalid abc passed to abc2cart")

    cart = []
    sin = np.sin(abc[1, :])
    cos = np.cos(abc[1, :])
    cart.append([sin[2], cos[2], 0.0])
    cart.append([0.0, 1.0, 0.0])
    cart.append([(cos[1]-cos[0]*cos[2])/sin[2], cos[0], 0.0])
    cart[2][2] = np.sqrt(1.0-cart[2][0]**2.0-cart[2][1]**2.0)
    cart = np.array([np.array(cart[i])*abc[0, i] for i in range(3)])

    return cart


def cart2abc(cart):
    """Transforms a Cartesian representation of lattice parameters
       into an axes and angles one

    """

    cart = np.array(cart, copy=False)

    if cart.shape != (3, 3):
        raise ValueError("Invalid cart passed to cart2abc")

    abc = np.zeros((2, 3))
    abc[0, :] = np.linalg.norm(cart, axis=1)
    # Now for the angles
    cart_roll_1 = np.roll(cart, 1, axis=0)
    cart_roll_2 = np.roll(cart, 2, axis=0)
    cross_cart = np.linalg.norm(np.cross(cart_roll_1, cart_roll_2), axis=1)
    dot_cart = np.diagonal(np.dot(cart_roll_1, cart_roll_2.T))
    abc[1, :] = np.arctan2(cross_cart, dot_cart)

    return abc


def hkl2d2_matgen(abc):
    """Generate a matrix that turns hkl indices into inverse crystal
       plane distances for a given lattice in ABC form

    """

    abc = np.array(abc, copy=False)

    if abc.shape != (2, 3):
        raise ValueError("Invalid abc passed to hkl2d2_matgen")

    sin = np.sin(abc[1, :])
    cos = np.cos(abc[1, :])
    a2b2 = (abc[0, 0]*abc[0, 1])**2.0
    a2c2 = (abc[0, 0]*abc[0, 2])**2.0
    b2c2 = (abc[0, 1]*abc[0, 2])**2.0
    abc_prod = np.prod(abc[0, :])

    hkl2d2 = np.array([[b2c2*sin[0]**2.0,
                        abc_prod*abc[0, 2]*(cos[0]*cos[1]-cos[2]),
                        abc_prod*abc[0, 1]*(cos[0]*cos[2]-cos[1])],
                       [abc_prod*abc[0, 2]*(cos[0]*cos[1]-cos[2]),
                        a2c2*sin[1]**2.0,
                        abc_prod*abc[0, 0]*(cos[1]*cos[2]-cos[0])],
                       [abc_prod*abc[0, 1]*(cos[0]*cos[2]-cos[1]),
                        abc_prod*abc[0, 0]*(cos[1]*cos[2]-cos[0]),
                        a2b2*sin[2]**2.0]])

    hkl2d2 /= abc_prod**2.0*(1.0-np.dot(cos, cos)+2.0*np.prod(cos))

    return hkl2d2


def inv_plane_dist(hkl, hkl2d2):
    """Calculate inverse planar distance for a given set of
       Miller indices h, k, l.

    """

    hkl = np.array(hkl, copy=False)

    if hkl.shape != (3,):
        raise ValueError("Invalid hkl passed to inv_plane_dist")

    return np.sqrt(np.dot(np.dot(hkl2d2, hkl), hkl)[0, 0])


def minimum_supcell(max_r, latt_cart=None, r_matrix=None,
                    pbc=[True, True, True]):
    """
    Generate the bounds for a supercell containing a sphere
    of given radius, knowing the unit cell.

    | Args:
    |   max_r (float): radius of the sphere contained in the supercell
    |   latt_cart (np.ndarray): unit cell in cartesian form
    |   r_matrix (np.ndarray): matrix for the quadratic form returning
    |                          r^2 for this supercell.
    |                          Alternative to latt_cart, for a direct
    |                          space cell would be equal to
    |                          np.dot(latt_cart, latt_cart.T)
    |   pbc ([bool, bool, bool]): periodic boundary conditions - if
    |                             a boundary is not periodic the
    |                             range returned will always be zero
    |                             in that dimension

    | Returns:
    |   shape (tuple[int]):  shape of the supercell to be built.

    | Raises:
    |   ValueError: if some of the arguments are invalid

    """

    if latt_cart is not None:
        latt_cart = np.array(latt_cart, copy=False)
        if latt_cart.shape != (3, 3):
            raise ValueError("Invalid latt_cart passed to minimum_supcell")
        r_matrix = np.dot(latt_cart, latt_cart.T)
    elif r_matrix is not None:
        r_matrix = np.array(r_matrix, copy=False)
        if r_matrix.shape != (3, 3):
            raise ValueError("Invalid r_matrix passed to minimum_supcell")
    else:
        raise ValueError("One between latt_cart and r_matrix has to"
                         "be present")

    # The logic of this algorithm isn't obvious, so let's make it clearer.
    #
    # What we are basically looking for is the AABB bounding box of a sphere
    # in absolute space.
    # This becomes a rotated ellipsoid in fractional coordinates space.
    # r_matrix represents a quadratic form which turns a fractional coordinate
    # into a squared distance in space.
    # By diagonalizing it we can calculate the transformation matrix
    # which morphs a simple unit sphere into this ellipsoid through
    # scaling (with the eigenvalues) and rotating (with the eigenvectors).
    # We have thus that a point p in unit sphere space becomes q = M*p
    # in fractional space. On the other hand, a point q in fractional space
    # will be transformed into absolute space by p = M^-1*q.
    # Now, the boundaries of the ellipsoids are none other but the points
    # which, in regular space, have normals aligned with the axes.
    # And NORMALS transform between spaces with a matrix that is the
    # inverse transpose of the regular one. So the logic goes:
    # - take an axis direction in fractional space (for example, [1,0,0])
    # - transform it into absolute space by premultiplying (M^-1)^-1^T = M^T
    # - now that we have the direction of the normal in absolute space,
    # we just need to normalize it to find the point p which has that normal
    # (thanks to the properties of the unit sphere)
    # - then we transform back to fractional space with M and there you go,
    # that's your boundary point.
    # - wash, rinse, repeat for all axes directions.
    # Find the maxima in each direction and you've got your box.

    r_evals, r_evecs = np.linalg.eigh(r_matrix)

    # Unit sphere - to - ellipsoid transformation matrix
    utransf_matrix = np.dot(r_evecs, np.diag(1.0/np.sqrt(r_evals)))
    # To find the boundaries, we need to iterate over the three main
    # directions. We do this implicitly though.
    qmatrix = np.dot(utransf_matrix,
                     max_r*(utransf_matrix/np.linalg.norm(utransf_matrix,
                                                          axis=1)[:, None]).T)
    r_bounds = np.max(np.ceil(abs(qmatrix)), axis=1).astype(int)
    r_bounds = np.where(pbc, r_bounds, 0)

    return tuple([2*r+1 for r in r_bounds])


def supcell_gridgen(latt_cart, shape):
    """
    Generate a full linearized grid for a supercell with r_bounds
    and a base unit cell in Cartesian form.

    | Args:
    |   latt_cart (np.ndarray): unit cell in cartesian form
    |   shape (tuple[int]):  shape of the supercell to be built,
    |                        as returned by minimum_supcell.

    | Returns:
    |   neigh_i_grid (np.ndarray): supercell grid in fractional coordinates
    |   neigh_grid (np.ndarray): supercell grid in cartesian coordinates

    | Raises:
    |   ValueError: if some of the arguments are invalid

    """

    latt_cart = np.array(latt_cart, copy=False)
    if latt_cart.shape != (3, 3):
        raise ValueError("Invalid latt_cart passed to supcell_gridgen")

    shape = np.array(shape, copy=False).astype(int)
    if shape.shape != (3,):
        raise ValueError("Invalid shape passed to supcell_gridgen")

    min_bounds = (-((shape-1)/2)).astype(int)
    max_bounds = shape+min_bounds
    x_range = range(min_bounds[0], max_bounds[0])
    y_range = range(min_bounds[1], max_bounds[1])
    z_range = range(min_bounds[2], max_bounds[2])

    # We now generate a grid of neighbours to check for contact with
    # First just the supercell indicess
    neigh_i_grid = np.swapaxes(np.meshgrid(x_range, y_range, z_range), 0, 3)
    neigh_i_grid = np.reshape(neigh_i_grid, (-1, 3))
    # Then the actual supercell cartesian shifts
    neigh_grid = np.dot(neigh_i_grid, latt_cart)

    return neigh_i_grid, neigh_grid


def minimum_periodic(v, latt_cart, exclude_self=False,
                     pbc=[True, True, True]):
    """
    Find the shortest periodic equivalent vector for a list of vectors and a
    given lattice.

    | Args:
    |   v (np.ndarray): list of 3-vectors representing points or vectors to
    |                   reduce to their closest periodic version
    |   latt_cart (np.ndarray): unit cell in cartesian form
    |   exclude_self (bool): if True, any vector that is equal to zero will be
    |                        excluded, and its closest non-zero periodic
    |                        version will be considered instead. Default is
    |                        False
    |   pbc ([bool, bool, bool]): periodic boundary conditions - if
    |                             a boundary is not periodic the
    |                             range returned will always be zero
    |                             in that dimension

    | Returns:
    |   v_period (np.ndarray): array with the same shape as v, containing the
    |                          vectors in periodic reduced form
    |   v_cells (np.ndarray): array of triples of ints, corresponding to the
    |                         cells from which the various periodic copies of
    |                         the vectors were taken. For an unchanged vector
    |                         will be all [0,0,0]

    """

    max_r = np.amax(np.linalg.norm(v, axis=-1))
    scell_shape = minimum_supcell(max_r, latt_cart, pbc=pbc)
    neigh_i_grid, neigh_grid = supcell_gridgen(latt_cart, scell_shape)
    v_period = np.array(v, copy=False)[:, None, :] + neigh_grid[None, :, :]
    v_norm = np.linalg.norm(v_period, axis=-1)
    if exclude_self:
        v_norm = np.where(v_norm > 0, v_norm, np.inf)
    min_copies = np.argmin(v_norm, axis=1)
    v_period = v_period[range(len(v)), min_copies, :]

    return v_period, neigh_i_grid[min_copies]


def all_periodic(v, latt_cart, max_r, pbc=[True, True, True]):
    """
    Find all the periodic equivalent vectors for a list of vectors and a
    given lattice falling within a given length.

    | Args:
    |   v (np.ndarray): list of 3-vectors representing points or vectors to
    |                   produce periodic versions of
    |   latt_cart (np.ndarray): unit cell in cartesian form
    |   max_r (float): maximum length of periodic copies of vectors
    |   pbc ([bool, bool, bool]): periodic boundary conditions - if
    |                             a boundary is not periodic the
    |                             range returned will always be zero
    |                             in that dimension

    | Returns:
    |   v_period (np.ndarray): array with the same shape as v, containing the
    |                          vectors in periodic reduced form
    |   v_index (np.ndarray): indices (referring to the original array v) of
    |                         the array of which the corresponding element of
    |                         v_period is a copy
    |   v_cells (np.ndarray): array of triples of ints, corresponding to the
    |                         cells from which the various periodic copies of
    |                         the vectors were taken. For an unchanged vector
    |                         will be all [0,0,0]

    """

    scell_shape = minimum_supcell(max_r, latt_cart, pbc=pbc)
    neigh_i_grid, neigh_grid = supcell_gridgen(latt_cart, scell_shape)
    v_period = np.array(v, copy=False)[:, None, :] + neigh_grid[None, :, :]
    r_copies = np.where(np.linalg.norm(v_period, axis=-1) <= max_r)
    v_period = v_period[r_copies[0], r_copies[1], :]

    return v_period, r_copies[0], neigh_i_grid[r_copies[1]]


def is_string(s):
    """Checks whether s is a string, with Python 2 and 3 compatibility"""
    try:
        return isinstance(s, basestring)
    except NameError:
        # It must be Python 3!
        return isinstance(s, str)


def safe_input(question):
    """Ask for user input with Python 2 and 3 compatibility"""
    try:
        return raw_input(question)
    except NameError:
        # Python 3
        return input(question)


def safe_communicate(subproc, stdin=''):
    """Executes a Popen.communicate and returns output in a way that is
    compatible with Python 2 & 3 keeping input and output as strings (since
    Python 3 requires bytes objects otherwise)"""
    if not subproc.universal_newlines:
        stdin = stdin.encode('utf-8') if hasattr(stdin, 'encode') else stdin
    stdout, stderr = map(lambda x: x.decode() if hasattr(x, 'decode') else x,
                         subproc.communicate(stdin))

    return stdout, stderr


# Inspecting arguments of a function, Python 2 and 3 way
if hasattr(inspect, 'signature'):
    def inspect_args(f):
        fsig = inspect.signature(f)
        args = fsig.parameters
        nargs = len(args)
        nargs_def = len([p for p in args
                         if args[p].default != inspect.Signature.empty])
        return (nargs, nargs_def)
else:
    def inspect_args(f):
        argspec = inspect.getargspec(f)
        return (len(argspec.args),
                (0 if argspec.defaults is None else len(argspec.defaults)))


# Importing a module from filename with compatibility across Python versions

def import_module(mpath):

    mname = seedname(mpath)

    # Python 3.5+ version
    try:
        import importlib.util
    except ImportError:
        pass
    else:
        spec = importlib.util.spec_from_file_location(mname, mpath)
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)
        return foo

    # Python 3.3-3.4 version
    try:
        from importlib.machinery import SourceFileLoader
    except ImportError:
        pass
    else:
        return SourceFileLoader(mname, mpath).load_module()

    # Python 2 version
    try:
        import imp
    except ImportError:
        # Ok, what the hell?
        raise ImportError('Could not import file ' + mpath)
    else:
        return imp.load_source(mname, mpath)

    # We should never get here, but...
    return None


def list_distance(l1, l2):
    """Return an integer distance between two lists (number of differing
    elements)"""

    ldiff = list(l2[:])
    d = 0
    d += len(l1)-len([ldiff.remove(el) for el in l1 if el in ldiff])
    d += len(ldiff)
    return d


def swing_twist_decomp(quat, axis):
    """Perform a Swing*Twist decomposition of a Quaternion. This splits the
    quaternion in two: one containing the rotation around axis (Twist), the
    other containing the rotation around a vector parallel to axis (Swing).

    Returns two quaternions: Swing, Twist.
    """

    # Current rotation axis
    ra = quat.q[1:]
    # Ensure that axis is normalised
    axis_norm = axis/np.linalg.norm(axis)
    # Projection of ra along the given axis
    p = np.dot(ra, axis_norm)*axis_norm
    # Create Twist
    qin = [quat.q[0], p[0], p[1], p[2]]
    twist = Quaternion(qin/np.linalg.norm(qin))
    # And Swing
    swing = quat*twist.conjugate()

    return swing, twist

### Clebsch-Gordan and Wigner-3j symbols ###


def clebsch_gordan(j, m, j1, m1, j2, m2):
    """ Clebsch-Gordan cohefficients for given quantum numbers:
    j, m, j1, m1, j2, m2

    The numbers passed can be arrays (just make sure they're the
    same size).
    """

    # Is it a single value?
    try:
        n = len(j)
    except TypeError:
        n = 1

    # Safety checks
    jm_all = np.array([j, m, j1, m1, j2, m2]).T
    if n == 1:
        jm_all = jm_all[None, :]
    elif len(jm_all.shape) == 1:
        # Inconsistent lenghts!
        raise ValueError('Not all arrays passed have the same size')
    if (((jm_all*2) % 1 != 0).any() or
            (np.abs(jm_all[:, 1::2]) > jm_all[:, ::2]).any()):
        raise ValueError('Invalid momentum values')

    j, m, j1, m1, j2, m2 = jm_all.T

    # Gotta do this the hard way...
    # Find all valid k

    kmax = np.min([j1+j2-j,          # j1+j2-j-k >= 0
                   j1-m1,            # j1-m1-k >= 0
                   j2+m2],           # j2+m2-k >= 0
                  axis=0)
    kmin = np.max([j*0,              # k >= 0
                   -j+j2-m1,         # j-j2+m1+k >= 0
                   -j+j1+m2],        # j-j1+m2+k >= 0
                  axis=0)

    ks = [np.arange(np.ceil(k0), np.floor(k1)+1)
          for k0, k1 in zip(kmin, kmax)]

    # Calculate k sum
    ksum = [np.sum((-1)**k/(factorial(k)*factorial(j1[i]+j2[i]-j[i]-k) *
                            factorial(j1[i]-m1[i]-k)*factorial(j2[i]+m2[i]-k) *
                            factorial(j[i]-j2[i]+m1[i]+k) *
                            factorial(j[i]-j1[i]-m2[i]+k)))
            for i, k in enumerate(ks)]

    cg = np.where(m == m1+m2,
                  np.sqrt((2*j+1)*factorial(j+j1-j2)*factorial(j-j1+j2) *
                          factorial(j1+j2-j)/factorial(j1+j2+j+1)) *
                  np.sqrt(factorial(j+m)*factorial(j-m)*factorial(j1-m1) *
                          factorial(j1+m1)*factorial(j2-m2)*factorial(j2+m2)) *
                  ksum,
                  0)
    if n == 1:
        cg = cg[0]

    return cg


def wigner_3j(j1, m1, j2, m2, j3, m3):
    """ Wigner 3j symbols for given quantum numbers:

    /                 \
    | j1    j2     j3 |
    | m1    m2     m3 |
    \                 /

    The numbers passed can be arrays (just make sure they're the
    same size).
    """

    # Expressed as a function of Clebsch-Gordan cohefficients
    j1, m1, j2, m2, j3, m3 = np.array([j1, m1, j2, m2, j3, m3])
    return (clebsch_gordan(j3, -m3, j1, m1, j2, m2) *
            (-1)**((j1-j2-m3) % 2)/np.sqrt(2*j3+1))


def max_distance_in_cell(cell):
    """Maximum distance between two points achievable inside a single unit
    cell.
    Relies on the fact that the isosurface for equal distance in cartesian
    space is an ellipsoid, meaning that the extreme has to be one of the 8
    corners of the cube of side 1 in fractional space.
    """

    corners = np.array(np.meshgrid(*[(-1, 1)]*3)).reshape((3, -1))
    # Upper distance bound in cell
    return np.amax(np.linalg.norm(np.dot(cell.T, corners), axis=0))


def periodic_bridson(cell, rmin, max_attempts=30,
                     prepoints=None, prepoints_cuts=None,
                     maxblock=1000):
    """ Periodic version of the Bridson algorithm for generation of Poisson
    sphere distributions of points. This returns a generator.

    | Args:
    |   cell (np.ndarray): periodic cell in which to create the points.
    |   rmin (float): minimum distance between each generated point.
    |   max_attempts (int): maximum number of candidate neighbours generated
    |                       for each point.
    |   prepoints (np.ndarray or list): pre-existing points to avoid during
    |                                   generation. These must be in
    |                                   fractional coordinates.
    |   prepoints_cuts (np.ndarray or list): custom cutoffs for each prepoint.
    |                                        If not included defaults to rmin.
    |   maxblock (int):     maximum size of a grid block to be processed in a
    |                       single function (used to avoid excess memory use)

    | Returns:
    |   bridsonGen (generator): an iterator producing Poisson-sphere like
    |                           distributed points in cell until space runs
    |                           out or enough attempts fail.
    """

    # 1. Compute the necessary number of divisions for the cell's grid
    ubound = max_distance_in_cell(cell)
    N = int(np.ceil(ubound/rmin))

    # 2. Mask for finding existing points that are too close
    grid_cell = cell/N
    shape = minimum_supcell(rmin, grid_cell)
    # Is it valid?
    if (np.array(shape) > N).any():
        raise ValueError('Value of rmin is too big for this unit cell')
    checkMask = np.array(np.meshgrid(*[range(int((1-i)/2), int((i-1)/2+1))
                                       for i in shape],
                                     indexing='ij')).reshape((3, -1))
    checkMask = np.array([m for m in checkMask.T if not (m == 0).all()]).T
    # 2.5 Mask for searching neighbours
    shape = minimum_supcell(2*rmin, grid_cell)
    newMask = np.array(np.meshgrid(*[range(int((1-i)/2), int((i-1)/2+1))
                                     for i in shape],
                                   indexing='ij')).reshape((3, -1))
    newMask = np.array([m for m in newMask.T if not (m == 0).all()]).T

    # 3. Now generate the grid
    grid = np.zeros((N, N, N)).astype(int)
    grid_points = np.zeros((N, N, N, 3))
    # 3.5 if there are prepoints, fill in the grid cells that are too close to
    # begin with
    if prepoints is not None:
        if prepoints_cuts is None:
            prepoints_cuts = np.ones(prepoints.shape[0])*rmin
        else:
            prepoints_cuts = np.array(prepoints_cuts)
        # Create corner points for cells
        grid_origins = np.array(np.meshgrid(*[range(N)]*3,
                                            indexing='ij')).reshape((3, -1))
        cell_corners = np.array(np.meshgrid(*[[0, 1]]*3, indexing='ij')
                                ).reshape((3, -1))
        for i0 in range(0, N**3, maxblock):
            go = grid_origins[:, i0:i0+maxblock]
            grid_corners = (go[:, :, None]+cell_corners[:, None, :])/N
            dfx = (prepoints[:, :, None, None] - grid_corners[None, :, :, :]
                   + 0.5) % 1-0.5
            r = np.linalg.norm(np.tensordot(dfx, cell, axes=(1, 0)), axis=-1)
            full = np.any(np.all(r < prepoints_cuts[:, None, None], axis=-1),
                          axis=0)
            grid[tuple(grid_origins[:, i0+np.where(full)[0]])] = 2

    # 4. And the queue with a random, non-occupied grid point
    if prepoints is None:
        queue = [np.array([0, 0, 0])]
    else:
        free_ijk = np.array(np.where(grid == 0))
        queue = [free_ijk[:, np.random.randint(free_ijk.shape[1])]]

    # Start iterations
    while np.sum(grid != 0) < N**3 and len(queue) > 0:
        # While there is still free space and queued points...
        iter_ijk0 = queue.pop()
        iter_mask = (newMask+iter_ijk0[:, None]) % N
        attempts = max_attempts

        # Try to generate points around the dequeued cell
        while attempts > 0:
            attempts -= 1
            candidates = np.array(iter_mask[:,
                                            np.where(grid[tuple(iter_mask)] ==
                                                     0)][:, 0])
            good = False
            if candidates.shape[1] == 0:
                continue
            i = np.random.randint(candidates.shape[1])
            ijk = candidates[:, i]
            # Pick the point
            fp = (ijk+np.random.random(3))/N
            # Check it
            near_mask = (checkMask+ijk[:, None]) % N
            near_points = tuple(near_mask[:, np.where(grid[tuple(near_mask)]
                                                      == 1)])
            near_points = grid_points[near_points][0]
            # Distances?
            if near_points.shape[0] > 0:
                dfx = (near_points - fp[None, :]+0.5) % 1-0.5
                r = np.linalg.norm(np.dot(dfx, cell), axis=1)
                good = not (r < rmin).any()
            else:
                good = True
            # Additional check: verify against prepoints
            if good and prepoints is not None:
                dfx = (prepoints - fp[None, :]+0.5) % 1-0.5
                r = np.linalg.norm(np.dot(dfx, cell), axis=1)
                good = not (r < prepoints_cuts).any()
            if good:
                grid[tuple(ijk)] = 1
                grid_points[tuple(ijk)] = fp
                queue.insert(0, ijk)
                yield np.dot(fp, cell)

    # So once we're here we ran out of options...
    return

# Function for creating labels for molecule sites


def recursive_mol_label(site_i, mol_indices, bonds, elems):
    """Creates a string for a given atom by traversing the molecular network
    starting with it.

    | Parameters:
    |   site_i (int): index of the atom for which the string must be
    |                 calculated
    |   mol_indices ([int]): indices of all atoms belonging to the molecule
    |   bonds (dict): dictionary of bonds within the molecule, containing
    |                 the indices of atoms as keys and lists of all the
    |                 indices of atoms they're bonded to as values.
    |   elems ([str]): list of element symbols for the entire system (must
    |                  be returned by ase.Atoms' get_chemical_symbols, NOT
    |                  follow the order the atoms appear in mol_indices)

    | Returns:
    |   recursive_label (str): a string representing the molecular network
    |                          traversed starting from site i
    """

    if site_i not in mol_indices:
        raise ValueError('Molecule does not contain given atom')

    def recursive_label(i, bonds, to_visit):
        # Remove from to_visit
        if i in to_visit:
            to_visit.remove(i)
        else:
            return None
        my_bonds = sorted([b for b in bonds[i] if b in to_visit])
        if len(my_bonds) > 0:
            bonded_label = sorted([recursive_label(j,
                                                   bonds,
                                                   to_visit)
                                   for j in my_bonds])
            bonded_label = [bl for bl in bonded_label if bl is not None]
            return '{0}[{1}]'.format(elems[i], ','.join(bonded_label))
        else:
            return '{0}'.format(elems[i])

    to_visit = list(mol_indices)

    return recursive_label(site_i, bonds, to_visit)


# Utilities for bonding graphs


@requireNetworkX('nx')
def get_bonding_graph(bond_mat, nx=None):
    return nx.Graph(bond_mat)


@requireNetworkX('nx')
def get_bonding_distance(bond_graph, i, j, nx=None):
    """Distance in bonds, returns -1 if the two are not connected"""
    try:
        d = nx.shortest_path_length(bond_graph, i, j)
    except nx.NetworkXNoPath:
        d = -1

    return d


# Utilities for clustering with scikit-learn

@requireScikitLearn('sk')
def get_sklearn_clusters(points, method, params, sk=None):
    """Cluster points using given method if found in sklearn.clusters module.
    Use params as a dictionary of parameters passed to the method."""

    try:
        __import__('sklearn.cluster')  # Avoids some weird ImportErrors. WTF.
        clustObj = getattr(sk, 'cluster').__dict__[method](**params)
    except KeyError:
        raise ValueError('Requested method is not present in scikit-learn')
    except TypeError:
        raise ValueError('Invalid parameters for method {0}'.format(method))

    return clustObj.fit_predict(points)

# Symmetric analysis utilities


@requireSpglib('spg')
def compute_asymmetric_distmat(struct, points, linearized=True,
                               return_images=False, images_centre=0,
                               symprec=1e-5, spg=None):
    """Given a symmetric structure, compute a distance matrix for the given
    fractional coordinate points which contains only the distances between 
    their closest symmetry equivalent sites in the asymmetric unit cell.

    Distances are computed in *fractional coordinates* space and are not 
    real distances. They do not necessarily map to real space distances
    either. Their purpose is merely to group points together by removing the
    effects of symmetry operations.

    | Parameters: 
    |   struct (ase.Atoms): the structure from which to extract the symmetry
    |                       operations
    |   points (np.ndarray): list of points, in fractional coordinates, to
    |                        compute the distance matrix for
    |   linearized (bool):  if True, returns only the linearised upper
    |                       triangle of the distance matrix, in the format
    |                       accepted by scipy.cluster.hierarchy.linkage.
    |                       Default is True
    |   return_images (bool): if True, return also a list of the closest
    |                         images for each point. The closest
    |                         symmetric image to images_centre is chosen.
    |   images_centre (int or np.ndarray): centre to use to compute closest
    |                                      images. If an integer, uses the
    |                                      given point of that index. If a 
    |                                      vector, uses the given coordinates.
    |                                      Should not be a high symmetry point.

    | Returns:
    |   distmat (np.ndarray): distance matrix for points
    |   images (np.ndarray): closest point images (only if return_images is
    |                        True)

    """

    points = np.array(points)
    N = points.shape[0]

    # Get symmetry operations
    symm = spg.get_symmetry_dataset(struct, symprec=symprec)
    rots = symm['rotations']
    transls = symm['translations']

    if linearized:
        distmat = np.zeros((N*(N-1))//2)
    else:
        distmat = np.zeros((N, N))
    all_images = (np.tensordot(rots, points,
                               axes=(2, 1))+transls[:, :, None]) % 1

    if return_images:
        closest_images = np.zeros((N, 3))
        if type(images_centre) is int:
            im0 = points[images_centre]
        else:
            im0 = np.array(images_centre)

        # Check im0
        im0_symm = (np.tensordot(rots, im0, axes=(2, 0))+transls) % 1

        if np.all(im0_symm[1:] == im0, axis=1).any():
            warnings.warn('Images centre is a high symmetry point, results may'
                          ' be incorrect', RuntimeWarning)

        df = (all_images-im0[None, :, None] + 0.5) % 1-0.5
        rf = np.linalg.norm(df, axis=1)
        minrf_i = np.argmin(rf, axis=0)
        closest_images = df[minrf_i, :, range(0, N)]+im0

    # Here we avoid full vectorisation to be safe against memory clutter.
    # Though it also means it's slower...
    for i in range(N-1):
        df = (all_images[:, :, i+1:]-all_images[None, 0, :, i, None]
              + 0.5) % 1-0.5
        rf = np.linalg.norm(df, axis=1)
        minrf_i = np.argmin(rf, axis=0)
        minrf = rf[minrf_i, range(rf.shape[1])]
        if linearized:
            distmat[i*(N-1)-(i*(i-1))//2:
                    (i+1)*(N-1)-(i*(i+1))//2] = minrf
        else:
            distmat[i, i+1:] = minrf
            distmat[i+1:, i] = distmat[i, i+1:]

    if return_images:
        return distmat, closest_images
    else:
        return distmat


# Repulsion algorithm to find the best place to add an atom


def rep_alg(v, iters=1000, attempts=10, step=1e-1, simtol=1e-5):
    """
    Repulsion algorithm, begins with a series of vectors v, finds a new one
    that is as far as possible, angle-wise, from all of them.
    The process is that of treating the tips of the vectors as particles on a
    sphere which repel each other. Of course, it's possible that multiple
    equilibrium positions exist, which is why multiple attempts can be made,
    and are judged identical or not based on a tolerance parameter

    | Parameters:
    |   v (np.ndarray): list of fixed vectors to avoid
    |   iters (int): number of iterations, default is 1000
    |   attempts (int): number of independent attempts with randomised starts,
    |                   default is 10
    |   step (float): step by which the 'particle' describing the vector will
    |                 be displaced at each iteration during the search.
    |                 Default is 1e-1.
    |   simtol (float): tolerance based on which two attempts are considered to
    |                   be effectively the same, checking if 1-v1.v2 < simtol.
    |                   Default is 1e-5.

    | Returns:
    |   out_v ([np.ndarray]): list of unit vectors of maximal distance from
    |                         the ones in v
    """

    # First, normalise the v vectors
    v = v/np.linalg.norm(v, axis=1)[:, None]

    out_v = np.zeros((0, 3))

    for a in range(attempts):

        # Random initialisation
        o_v = np.random.random(3)-0.5
        o_v /= np.linalg.norm(o_v)

        # Iterate
        for i in range(iters):
            F = o_v-v
            F /= np.linalg.norm(F, axis=1)[:, None]
            do_v = np.sum(F, axis=0)
            do_v *= step/np.linalg.norm(do_v)
            o_v += do_v
            o_v /= np.linalg.norm(o_v)

        # Check for similarities
        sim = np.dot(out_v, o_v)
        if not np.any(abs(1.0-sim) < simtol):
            out_v = np.concatenate((out_v, o_v[None, :]), axis=0)

    return out_v


def graph_specsort(L):
    # Spectral sorting of a graph from its Laplacian matrix

    evals, evecs = np.linalg.eigh(L)
    # Fiedler vector
    fied = evecs[:, 1]
    # Sign convention: more positive than negative components
    n = len(evals)
    fied *= -1 if np.sum(fied > 0) < n/2.0 else 1

    return np.argsort(fied)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Van der Waals radii

Available sets:

    - csd:  extrapolated from the Cambridge Structural Database by Santiago
    Alvarez, as seen in S. Alvarez, "A cartography of the van der Waals
    territories", Dalton Trans. 42, 8617 (2013)
    - jmol: extracted from the source code of JMol
    - ase:  default set for the Atomic Simulation Environment
    
    
"""

import json
import pkgutil
import numpy as np
from ase.data import atomic_numbers
from ase.data.vdw import vdw_radii as _vdw_radii_ase


def _load_vdw(name):

    _vdw_data = pkgutil.get_data('soprano', 'data/vdw_{0}.json'.format(name)
                                 ).decode('utf-8')
    _vdw_radii = np.array(json.loads(_vdw_data))

    return _vdw_radii


vdw_radii = {
    'ase': _vdw_radii_ase,
    'jmol': _load_vdw('jmol'),
    'csd': _load_vdw('csd')
}


def vdw_radius(el, vdwset='csd'):
    """Return Van der Waals radius for a certain element

    Return a Van der Waals radius for a given element, as taken from one
    of three available databases:

    - csd:  extrapolated from the Cambridge Structural Database by Santiago
    Alvarez, as seen in S. Alvarez, "A cartography of the van der Waals
    territories", Dalton Trans. 42, 8617 (2013) [default]
    - jmol: extracted from the source code of JMol
    - ase:  default set for the Atomic Simulation Environment


    | Args:
    |   el (str):       element symbol
    |   vdwset (str):   VdW set to use. Default is 'csd'.

    | Returns:
    |   vdw_radius (float): the Van der Waals radius
    """

    try:
        Z = atomic_numbers[el]
    except KeyError:
        raise ValueError('Invalid element symbol')

    return vdw_radii[vdwset][Z]



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Classes and functions required for processes that automatically submit jobs
to a queueing system working in the background.

These can be launched interactively from the command line. In order to do
that:

1. write your own implementation of a submitter class by inheriting from
   soprano.hpc.submitter.Submitter or use one of the provided ones;
2. write an input file in which you simply create an instance of said class
   and set up its parameters (ideally by calling the set_parameters method);
3. launch that submitter from the command line with the following command:

    ``python -m soprano.hpc.submitter start <filename>``

You can have multiple submitter instances, even of different types, defined
in the same file: in that case you will need to use the -n option to specify
which one you want to launch (the name you need to use is the name of the
*variable* you stored the instance in). If you are working on remote login and
you want to prevent the submitter from being terminated upon exiting your
session use the -nohup option.
To list which submitters from a given file are running, and how long have they
been running for, just use:

    ``python -m soprano.hpc.submitter list <filename>``

Similarly, you can stop a running submitter with:

    ``python -m soprano.hpc.submitter stop <filename>``

Submitters have a 'name' property and will save a <name>.log file in which any
output from their run can be stored.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import

from soprano.hpc.submitter.queues import QueueInterface
from soprano.hpc.submitter.submit import Submitter
from soprano.hpc.submitter.castep import CastepSubmitter



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility functions to compare clusterings and evaluate similarity"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np


def confmat(clust1, clust2):
    """Calculate the confusion matrix for two clusterings of the same
    collection. The confusion matrix is defined so that element i,j contains
    the number of elements that are in common between cluster i of the first
    clustering and cluster j of the second:

    M[i,j] = len(set(c[i]).intersection(c[j]))

    | Args:
    |   clust1 (list): a series of clusters in the form of a list of slices,
    |                  like the second value returned by one of the clustering
    |                  methods in PhylogenCluster.
    |   clust2 (list): same as above.

    | Returns:
    |   confmat (np.ndarray): confusion matrix for the two clusterings.
    """

    cmat = [[len(set(c1).intersection(c2)) for c2 in clust2]
            for c1 in clust1]

    return np.array(cmat)


def norm_confmat(clust1, clust2):
    """Calculate the normalised confusion matrix for two clusterings of the
    same collection. The confusion matrix is defined as in the docstring of
    confmat. For the normalisation, each element i,j is divided by the
    geometric mean of the sizes of cluster i of the first clustering and
    cluster j of the second:

    NM[i,j] = M[i,j]/(len(c[i])*len(c[j]))**0.5

    | Args:
    |   clust1 (list): a series of clusters in the form of a list of slices,
    |                  like the second value returned by one of the clustering
    |                  methods in PhylogenCluster.
    |   clust2 (list): same as above.


    | Returns:
    |   nconfmat (np.ndarray): normalised confusion matrix for the two
    |                          clusterings.
    """

    cmat = confmat(clust1, clust2)
    norm = [[(len(c1)*len(c2))**0.5 for c2 in clust2] for c1 in clust1]

    return cmat/np.array(norm)


def fowles_mallows_index(clust1, clust2):
    """Calculate the Fowles-Mallows index, a measure of similarity between two
    clusterings defined as:

    F = (WI*WII)**0.5

    with

    WI = N11/sum_k(n_k*(n_k-1)/2)
    WII = N11/sum_k(n'_k*(n'_k-1)/2)

    with N11 being the number of pairs of points that are in the same cluster
    in both clusterings, and n_k (n'_k) the number of elements in cluster k
    of the first (second) clustering.

    Ref: 
    Fowlkes, E. B.; Mallows, C. L. (1 September 1983). 
    "A Method for Comparing Two Hierarchical Clusterings". 
    Journal of the American Statistical Association. 78 (383): 553. 
    doi:10.2307/2288117

    | Args:
    |   clust1 (list): a series of clusters in the form of a list of slices,
    |                  like the second value returned by one of the clustering
    |                  methods in PhylogenCluster.
    |   clust2 (list): same as above.


    | Returns:
    |   fm_ind (float): the Fowles-Mallows index 
    """

    cmat = confmat(clust1, clust2)

    N11 = (np.sum(cmat**2.0)-np.sum(cmat))/2.0
    n1 = np.array([len(c) for c in clust1])
    n2 = np.array([len(c) for c in clust2])
    WI = N11/np.sum(0.5*n1*(n1-1))
    WII = N11/np.sum(0.5*n2*(n2-1))

    return (WI*WII)**0.5



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generator producing structures with a randomly positioned point defect"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import json
import pkgutil
import itertools
import numpy as np
from ase import Atoms
from ase.data import atomic_numbers
# Internal imports
import soprano.utils as utils
from soprano.selection import AtomSelection
from soprano.properties.linkage import Bonds
from soprano.data import vdw_radii


def defectGen(struct, defect, poisson_r=None, avoid_atoms=True,
              vdw_set='csd', vdw_scale=1, max_attempts=30):
    """Generator function to create multiple structures with a defect of a
    given element randomly added to the existing cell. The defects can be
    distributed following a random uniform distribution or a Poisson-sphere
    distribution which guarantees that no two defects are created closer than
    a certain distance. These are created using a variation on Bridson's
    algorithm. For that reason the defects will be created contiguously and
    added in succession. A full cover of all the empty space will be provided
    only by extracting structures until StopIteration is raised.

    | Args:
    |   struct (ase.Atoms): the starting structure. All defects will be added
    |                       to it.
    |   defect (str): element symbol of the defect to add.
    |   poisson_r (float): if this argument is present, the Poisson sphere
    |                      distribution rather than random uniform will be
    |                      used, and this will be the minimum distance in
    |                      Angstroms between the generated defects.
    |   avoid_atoms (bool): if True, defects too close to the existing atoms
    |                       in the cell will be rejected. The cutoff distance
    |                       will be estimated based on the Van der Waals radii
    |                       set of choice. Default is True.
    |   vdw_set({ase, jmolm csd}): set of Van der Waals radii to use. Only 
    |                              relevant if avoid_atoms is True. 
    |                              Default is csd [S. Alvarez, 2013].
    |   vdw_scale (float): scaling factor to apply to the base Van der Waals
    |                      radii values. Only relevant if avoid_atoms is True.
    |                      Values bigger than one make for larger empty
    |                      volumes around existing atoms.
    |   max_attempts(int): maximum number of attempts used to generate a
    |                      random point. When using a uniform distribution,
    |                      this is the maximum number of attempts that will be
    |                      done to avoid existing atoms (if avoid_atoms is
    |                      False, no attempts are needed). When using a Poisson
    |                      sphere distribution, this is a parameter of the
    |                      Bridson-like algorithm and will include also
    |                      avoiding previously generated defects. Default
    |                      is 30.

    | Returns:
    |   defectGenerator (generator): an iterator object that yields
    |                                structures with randomly distributed
    |                                defects.
    """

    # Lattice parameters
    cell = struct.get_cell()

    # Van der Waals radii (if needed)
    if avoid_atoms:
        avoid_fpos = struct.get_scaled_positions()
        avoid_vdw = vdw_radii[vdw_set][struct.get_atomic_numbers()]
        avoid_cut = (avoid_vdw +
                     vdw_radii[vdw_set][atomic_numbers[defect]]
                     )*vdw_scale/2.0

    if poisson_r is None:
        # Random generation
        while True:
            good = False
            attempts = max_attempts

            while not good and attempts > 0:
                # 1. Generate point in fractional coordinates
                fp = np.random.random((3,))
                # 2. If required check it against existing atoms
                if avoid_atoms:
                    dx = np.dot(avoid_fpos-fp[None, :], cell)
                    dr = np.linalg.norm(dx, axis=1)
                    good = (dr > avoid_cut).all()
                else:
                    good = True
                attempts -= 1

            if good:
                yield Atoms(defect, scaled_positions=[fp], cell=cell)+struct
            else:
                raise RuntimeError('Failed to generate a defect obeying'
                                   ' constraints with the given number of '
                                   'attempts.')
    else:
        # Use Bridson's algorithm
        if avoid_atoms:
            gen = utils.periodic_bridson(cell, poisson_r,
                                         max_attempts=max_attempts,
                                         prepoints=avoid_fpos,
                                         prepoints_cuts=avoid_cut)
        else:
            gen = utils.periodic_bridson(cell, poisson_r,
                                         max_attempts=max_attempts)

        while True:
            p = next(gen, None)
            if p is None:
                return
            yield Atoms(defect, positions=[p], cell=cell) + struct


def substitutionGen(struct, subst, to_replace=None, n=1, accept=None):
    """Generator function to create multiple structures with a defect of a
    given element substituted in the existing cell. The defects will be put in
    place of the atoms passed in the to_replace selection. If none is passed,
    all atoms will be replaced in turn. Multiple defects can be included, in
    which case all permutations will be generated.
    It is also possible to reject some configurations based on the output of a
    filter function.

    | Args:
    |   struct (ase.Atoms): the starting structure. All defects will be added
    |                       to it.
    |   subst (str): element symbol of the defect to add.
    |   to_replace (AtomSelection): if present, only atoms belonging to this
    |                               selection will be substituted.
    |   n (int): number of defects to include in each structure. Default is 1.
    |   accept (function): a function that determines whether a generated
    |                      structure should be accepted or rejected. Takes as
    |                      input the generated structure and a tuple of the
    |                      indices of the substituted atoms, and must return a
    |                      bool. If False, the structure will be rejected.
    | Returns:
    |   defectGenerator (generator): an iterator object that yields
    |                                structures with all possible
    |                                substitutions.
    """

    if to_replace is None:
        to_replace = AtomSelection.all(struct)

    defconfs = itertools.combinations(to_replace.indices, n)
    elems = np.array(struct.get_chemical_symbols()).astype('<U2')

    for dc in defconfs:
        dstruct = struct.copy()
        delems = elems.copy()
        delems[list(dc)] = subst
        dstruct.set_chemical_symbols(delems)

        if accept is not None:
            if not accept(dstruct, dc):
                continue

        yield dstruct


def additionGen(struct, add, to_addition=None, n=1, add_r=1.2,
                accept=None):
    """Generator function to create multiple structures with an atom of a
    given element added in the existing cell. The atoms will be attached to 
    the atoms passed in the to_addition selection. If none is passed,
    all atoms will be additioned in turn. Multiple defects can be included, in
    which case all permutations will be generated. The algorithm will try
    adding the atom in the direction that seems most compatible with all the
    already existing bonds. If multiple directions satisfy the condition, they
    will all be tested.
    It is also possible to reject some configurations based on the output of a
    filter function.

    | Args:
    |   struct (ase.Atoms): the starting structure. All atoms will be added
    |                       to it.
    |   add (str): element symbol of the atom to add.
    |   to_addition (AtomSelection): if present, only atoms belonging to this
    |                                selection will get an addition.
    |   n (int): number of new atoms to include in each structure. Default
    |            is 1.
    |   add_r (float): distance, in Angstroms, at which to add the atoms.
    |                  Default is 1.2 Ang
    |   accept (function): a function that determines whether a generated
    |                      structure should be accepted or rejected. Takes as
    |                      input the generated structure and a tuple of
    |                      the indices of the atoms to which the new atoms
    |                      were added, and must return a bool. The newly added
    |                      atoms will always be the last n of the structure.
    |                      If False, the structure will be rejected.
    | Returns:
    |   defectGenerator (generator): an iterator object that yields
    |                                structures with all possible additions.
    """

    if to_addition is None:
        to_addition = AtomSelection.all(struct)

    # Compute bonds
    bonds = Bonds.get(struct)

    cell = struct.get_cell()
    pos = struct.get_positions()

    # Separate bonds by atoms
    atom_bonds = [[] if i in to_addition.indices else None
                  for i in range(len(struct))]
    for b in bonds:
        v = pos[b[1]]-pos[b[0]]+np.dot(b[2], cell)
        try:
            atom_bonds[b[0]].append(v.copy())
        except AttributeError:
            pass
        try:
            atom_bonds[b[1]].append(-v.copy())
        except AttributeError:
            pass

    # Compute possible attachment points for each atom
    attach_v = [None]*len(struct)
    for i, bset in enumerate(atom_bonds):
        if bset is None:
            continue
        if len(bset) == 0:
            rndv = np.random.random((1, 3))-0.5
            rndv /= np.linalg.norm(rndv, axis=1)[:, None]
            attach_v[i] = rndv
        else:
            attach_v[i] = utils.rep_alg(bset)
    attach_v = np.array(attach_v)

    addconfs = itertools.combinations(to_addition.indices, n)

    for ac in addconfs:
        addpos = itertools.product(*attach_v[list(ac)])
        for ap in addpos:
            astruct = struct.copy()
            astruct += Atoms(add*n, positions=pos[list(ac)]
                             + np.array(ap)*add_r)

            if accept is not None:
                if not accept(astruct, ac):
                    continue

            yield astruct



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Utilities for remote submission, especially using Paramiko for SSH
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

try:
    import paramiko as pmk
except ImportError:
    pmk = None
import os
import glob
import fnmatch
from soprano.utils import is_string


class RemoteTarget(object):

    """RemoteTarget object

    Uses Paramiko to embed into one single object all we need to connect
    and send orders to some remote machine. It will accept a single host name
    as an argument and requires it to be present in the ~/.ssh/config file
    with all relevant parameters. It also must have an ssh key configurated
    for access (passwords are NOT accepted as they can't be stored and passed
    safely) and must have already been added to known_hosts (in other words,
    you must already have connected to it from the shell).

    """

    def __init__(self, host, timeout=1.0):
        """Initialize the RemoteTarget

        | Args:
        |   host (str): the name of the target to connect to, as present
        |               in the ~/.ssh/config file
        |   timeout (Optional[float]): connection timeout in seconds (default
        |                              is 1 second)

        """

        # Check that Paramiko is even present
        if pmk is None:
            raise RuntimeError('Paramiko not installed - RemoteTarget can not'
                               ' be initialised')

        # Ok, check that the hostname exists
        config = pmk.SSHConfig()
        config.parse(open(os.path.expanduser('~/.ssh/config')))

        if host not in config.get_hostnames():
            raise ValueError('Host '
                             '{0} not found in ~/.ssh/config'.format(host))

        hostdata = config.lookup(host)
        self._connect_args = {
            'hostname': hostdata['hostname'],
            'timeout': timeout
        }

        if 'port' in hostdata:
            self._connect_args['port'] = int(hostdata['port'])
        if 'user' in hostdata:
            self._connect_args['username'] = hostdata['user']
        if 'identityfile' in hostdata:
            self._connect_args['key_filename'] = hostdata['identityfile'][-1]

        # Now create the SSHClient and parse the system's hostkeys
        self._client = pmk.SSHClient()
        self._client.load_system_host_keys()

        # We're all ready now!

    @property
    def context(self):
        """Returns a context using this RemoteTarget to connect.
        This is done this way so that we only need to instantiate the Target
        once to avoid pointless overhead but still can handle the connection
        securely by making sure it's closed no matter what.
        """
        return RemoteTargetContext(self)


def _ensure_open_sftp(f):
    """Decorator for SFTP related functions in RemoteTargetContext"""

    def sftp_checked(self, *args, **kwargs):
        if self._sftp is None:
            self._sftp = self._client.open_sftp()
        return f(self, *args, **kwargs)

    return sftp_checked


class RemoteTargetContext(object):

    """RemoteTargetContext object

    Works as a context to be used with the 'with' statement. Should usually
    just be created by a RemoteTarget through the appropriate property.

    """

    def __init__(self, rT):
        self._client = rT._client
        self._connect_args = rT._connect_args
        self._sftp = None

    def __enter__(self):
        # Open the connection
        self._client.connect(**self._connect_args)
        return self

    def __exit__(self, type, value, traceback):
        # Make sure the connection is closed
        if self._sftp is not None:
            self._sftp.close()
        self._client.close()

    def run_cmd(self, cmd, cwd=None, stdin=None):
        """
        Run a command on the remote machine.

        | Args:
        |   cmd (str): command to run remotely
        |   cwd (Optional[str]): working directory in which to run the command
        |   stdin (Optional[str]): content to communicate to the command's
        |                          stdin

        | Returns:
        |   stdout, stderr (str): string representations of output and error
        |                         messages from the command

        """

        if cwd is not None:
            cmd = 'cd {0}; '.format(cwd) + cmd

        (_stdin,
         _stdout,
         _stderr) = self._client.exec_command(cmd,
                                              timeout=self._connect_args[
                                                  'timeout'])

        if stdin is not None:
            _stdin.write(stdin)
            _stdin.channel.shutdown_write()

        return _stdout.read(), _stderr.read()

    @_ensure_open_sftp
    def put_files(self, localpaths, remotedir):
        """
        Copy files to the remote machine via SFTP.

        | Args:
        |   localpaths (str or [str]): path of file(s) to copy. Can include
        |                              wildcards.
        |   remotedir (str): remote directory to copy the file(s) into.

        """

        if is_string(localpaths):
            localpaths = [localpaths]

        for lpath in localpaths:
            files = glob.glob(lpath)

            for f in files:
                _, fname = os.path.split(f)
                self._sftp.put(f, os.path.join(remotedir, fname),
                               confirm=True)

    @_ensure_open_sftp
    def get_files(self, remotepaths, localdir):
        """
        Download files from the remote machine via SFTP.

        | Args:
        |   remotepaths (str or [str]): path of file(s) to copy. Can include
        |                               wildcards.
        |   localdir (str): local directory to copy the file(s) into.

        """

        if is_string(remotepaths):
            remotepaths = [remotepaths]

        for rpath in remotepaths:
            remotedir, remotefiles = os.path.split(rpath)

            all_files = self._sftp.listdir(remotedir)
            files = fnmatch.filter(all_files, remotefiles)

            for f in files:
                _, fname = os.path.split(f)
                self._sftp.get(os.path.join(remotedir, f),
                               os.path.join(localdir, fname))



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of Submitter class

Base class for all Submitters to inherit from.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import time
import pickle
import shutil
import signal
import tempfile
import numpy as np
import threading as thr
import subprocess as sp
from datetime import datetime
from socket import timeout as TimeoutError

from soprano.utils import is_string, safe_communicate
from soprano.hpc.submitter import QueueInterface


class Submitter(object):

    """Submitter object

    Template to derive all specialised Submitters. These are meant to generate,
    submit and post-process any number of jobs on a queueing system in the form
    of a background process running on a head node. It implements
    methods that should be mostly overridden by the child classes.
    The following methods define its core behaviour:

    1) next_job is the function that outputs the specification for each new job
       to submit. The specification should be a dict with two members, 'name'
       (a string) and 'args' (ideally a dict). If no more jobs are available
       it should return None;
    2) setup_job takes as arguments name, args and folder (a temporary one
       created independently) and is supposed to generate the input files
       for the job before submission. It returns a boolean, confirming that
       the setup went well; if False, the job will be skipped;
    3) check_job takes as arguments job ID, name, args and folder and should
       return a bool confirmation of whether the job has finished or not. By
       default it simply checks whether the job is still listed in the queue,
       however other checks can be implemented in its place;
    4) finish_job takes as arguments name, args and folder and takes care of
       the post processing once a job is complete. Here meaningful data should
       be extracted and useful files copied to permament locations, as the
       temporary folder will be deleted immediately afterwards. It returns
       nothing;
    5) start_run takes no arguments, executes at the beginning of a run;
    6) finish_run takes no arguments, executes at the end of a run.
    7) save_state takes no arguments, returns a dict. It is executed when
       continuation=True is used and a run terminates. It will allow the user
       to add class-specific data to the dictionary that is stored in the
       pickle file (in addition to the default, namely the list and info on
       currently running jobs). This should be used for example to store state
       information that is necessary for job generation. It should be composed
       of serialisable objects.
    8) load_state takes as arguments the loaded data in dictionary form. It
       should perform the reverse operation of save_state, grabbing the info
       and restoring the Submitter's state to its previous condition.

    In addition, the Submitter takes a template launching script which can
    be tagged with keywords, mainly <name> for the job name or any other
    arguments present in args. These will be replaced with the appropriate
    values when the script is submitted.
    """

    def __init__(self, name, queue, submit_script, max_jobs=4, check_time=10,
                 max_time=3600, temp_folder=None, remote_workdir=None,
                 remote_getfiles=['*.*'], ssh_timeout=1.0,
                 continuation=False):
        """Initialize the Submitter object

        | Args:
        |   name (str): name to be used for this Submitter (two Submitters
        |               with the same name can't be launched in the same
        |               working directory)
        |   queue (QueueInterface): object describing the properties of the
        |                           interface to the queue system in use
        |   submit_script (str): text of the script to use when submitting a
        |                        job to the queue. All tags of the form <name>
        |                        will be replaced with the job's name, and all
        |                        similar tags of the form <[arg]> will be
        |                        replaced if the argument name is present in
        |                        the job's args dictionary
        |   max_jobs (Optional[int]): maximum number of jobs to submit at a
        |                             given time. Default is 4
        |   check_time (Optional[float]): time in seconds between consecutive
        |                                 checks for the queue status and
        |                                 attempts to submit new jobs. Default
        |                                 is 10
        |   max_time (Optional[float]): time in seconds the Submitter will run
        |                               for before shutting down. If set to
        |                               zero the thread won't stop until
        |                               killed with Submitter.stop.
        |   temp_folder (Optional[str]): where to store the temporary folders
        |                                for the calculations. By default it's
        |                                the current folder.
        |   remote_workdir (Optional[str]): if present, uses a directory on a
        |                                   remote machine by logging in via
        |                                   SSH. Must be in the format
        |                                   <host>:<path/to/directory>.
        |                                   Host must be defined in the user's
        |                                   ~/.ssh/config file - check the
        |                                   docs for RemoteTarget for more
        |                                   information. It is possible to
        |                                   omit the colon and directory, that
        |                                   will use the home directory of the
        |                                   given folder; that is HEAVILY
        |                                   DISCOURAGED though. Best practice
        |                                   would be to create an empty
        |                                   directory on the remote machine
        |                                   and use that, to avoid accidental
        |                                   overwriting/deleting of important
        |                                   files.
        |   remote_getfiles (Optional[list(str)]): list of files to be
        |                                          downloaded from the remote
        |                                          copy of the job's temporary
        |                                          directory. By default, all
        |                                          of them. Can be a list
        |                                          using specific names,
        |                                          wildcards etc. Filenames
        |                                          can also use the
        |                                          placeholder {name} to
        |                                          signify the job name, as
        |                                          well as any other element
        |                                          from the arguments.
        |   ssh_timeout (Optional[float]): connection timeout in seconds
        |                                  (default is 1 second)
        |   continuation (Optional[bool]): if True, when the Submitter is
        |                                  stopped it will not terminate the
        |                                  current jobs; rather, it will store
        |                                  the list in a pickle file.
        |                                  If the submitter is ran from the
        |                                  same folder then it will "pick up
        |                                  from where it left" and try
        |                                  recovering those jobs, then
        |                                  restart. If one wishes for
        |                                  additional values to be saved and
        |                                  restored, the save_state and
        |                                  load_state methods need to be
        |                                  defined.

        """

        # Check type
        if not isinstance(queue, QueueInterface):
            raise TypeError('A QueueInterface must be passed to the '
                            'Submitter')

        if not is_string(submit_script):
            raise TypeError('submit_script must be a string')

        self.name = name
        self.queue = queue
        self.submit_script = submit_script
        self.max_jobs = max_jobs
        self.check_time = check_time
        self.max_time = max_time if max_time > 0 else np.inf
        self.tmp_dir = (os.path.abspath(temp_folder)
                        if temp_folder is not None else '')

        # User defined signals
        self._free_signals = [signal.__dict__[s] for s in ('SIGUSR1',
                                                           'SIGUSR2')
                              if s in signal.__dict__]
        self._user_signals = {}

        self._log = None  # Will keep track of failed jobs etc.

        # Remote directory?
        if remote_workdir is None:
            self.host = None
        else:
            if ':' in remote_workdir:
                self.host, self.hostdir = remote_workdir.split(':', 1)
            else:
                self.host = remote_workdir
                self.hostdir = ''

        self.remote_getfiles = remote_getfiles

        self.queue.set_remote_host(self.host, ssh_timeout)

        self.continuation = continuation

    def set_parameters(self):
        """Set additional parameters. In this generic example class it has
        no arguments, but in specific implementations it will be used to
        add more variables without overriding __init__."""

        pass

    def add_signal(self, command, callback):
        """Add a signal listener to this submitter. Unix systems only allow
        for up to TWO user-defined signals to be specified.

        | Args:
        |   command (str): command that should be used to call this signal.
        |                  This would be used as:
        |                  python -m soprano.hpc.submitter <command> <file>
        |                  and will trigger the callback's execution
        |   callback (function<self> => None): method of the user defined
        |                      Submitter class to use as a callback when the
        |                      given signal is sent. Should accept and return
        |                      nothing.

        """

        # Is the command a reserved word?
        if command in ('start', 'stop', 'list'):
            raise ValueError('The commands start, stop and list are reserved'
                             ' and can not be used for custom signals.')

        # Are there any free signals left?
        try:
            signum = self._free_signals.pop()
        except IndexError:
            raise RuntimeError('Impossible to assign more signals - maximum '
                               'number supported by the OS has been reached.')

        self._user_signals[command] = [signum, callback]

    def remove_signal(self, command):
        """Remove a previously defined custom signal by its assigned command.

        | Args:
        |   command (str): command assigned to the signal handler to remove.

        """

        try:
            signum, _ = self._user_signals[command]
            del(self._user_signals[command])
            self._free_signals.append(signum)
        except KeyError:
            raise ValueError('Command does not correspon to any assigned'
                             ' signal.')

    def start(self):

        self._log = open(self.name + '.log', 'w')
        self.log('Starting run on {0}\n'.format(datetime.now()))

        self._jobs = {}
        self._waiting_jobs = []  # Jobs created but not yet submitted
        self._completed_jobs = []  # Jobs completed but not yet finalised

        if self.continuation and os.path.exists(self._pklname):
            self._load()

        self._running = True
        self._t0 = time.time()  # Starting time. Second precision is fine

        # Initialise signal catching
        signal.signal(signal.SIGINT, self._catch_signal)
        signal.signal(signal.SIGTERM, self._catch_signal)
        # Now assign the user-defined ones
        for cmd in self._user_signals:
            signum, cback = self._user_signals[cmd]
            signal.signal(signum, self._catch_signal)

        # Just a sanity check
        has_name = False
        for l in self.submit_script.split('\n'):
            if '<name>' in l.split('#')[0]:
                has_name = True
                break

        if not has_name:
            self.log('WARNING: the submission script does not contain the '
                     '<name> tag in any non-commented lines. '
                     'This is most likely an error - check your '
                     'input files\n')

        self.start_run()
        self._main_loop()
        self.finish_run()
        self.log('Run finished on {0}\n'.format(datetime.now()))
        self._log.close()

    def _catch_signal(self, signum, frame):
        if signum in (signal.SIGINT, signal.SIGTERM):
            # This catches the signal when termination is asked
            self.log('SIGTERM received - '
                     'Starting termination of this run...\n')
            self._terminate()
        else:
            for cmd in self._user_signals:
                if signum == self._user_signals[cmd][0]:
                    self._user_signals[cmd][1]()

    def _terminate(self):
        self._running = False
        # Also, kill all jobs still running
        if not self.continuation:
            for job_id in self._jobs.keys():
                self.queue.kill(job_id)
                # If needed, get the files from remote host
                if self.host is not None:
                    self._getjob_remote(self._jobs[job_id])
                self.finish_job(**self._jobs[job_id])
                shutil.rmtree(self._jobs[job_id]['folder'])
        else:
            self._save()

        self._jobs = {}

    def _main_loop(self):
        """Main loop run as separate thread. Should not be edited when
        inheriting from the class"""

        while self._running and (time.time()-self._t0) < self.max_time:

            loop_t0 = time.time()

            # Submit jobs
            while len(self._jobs) < self.max_jobs:

                if len(self._waiting_jobs) == 0:
                    njob = self.next_job()
                    if njob is None:
                        break
                    # Create the temporary folder
                    njob['folder'] = tempfile.mkdtemp(dir=self.tmp_dir)
                    # Perform setup
                    if not self.setup_job(**njob):
                        self.log('Job {0} did not pass setup check, '
                                 'skipping\n'.format(njob['name']))
                        # Remove the temporary directory
                        shutil.rmtree(njob['folder'])
                        self.log(('Folder {0} '
                                  'deleted\n').format(njob['folder']))
                        continue
                else:
                    njob = self._waiting_jobs.pop(0)

                # If we're working with a remote host, we need to copy the
                # input folder!
                if self.host is not None:
                    try:
                        self._putjob_remote(njob)
                    except TimeoutError:
                        self.log('Timeout when trying to push input files to'
                                 ' host. If it happens too frequently, try '
                                 'increasing the ssh_timeout argument.\n')
                        self._waiting_jobs.append(njob)
                        break

                # Create custom script
                job_script = self.submit_script.replace('<name>',
                                                        njob['name'])
                # Replace the rest of the tags
                for tag in njob['args']:
                    job_script = job_script.replace('<{0}>'.format(tag),
                                                    str(njob['args'][tag]))
                job_script = job_script.replace('<folder>',
                                                njob['folder'])

                # And submit! [Only if still running]
                if not self._running:
                    break
                else:
                    self.log('Submitting job '
                             '{0} to queue\n'.format(njob['name']))

                    if self.host is not None:
                        cwd = os.path.join(self.hostdir, njob['name'])
                    else:
                        cwd = njob['folder']

                    job_id = self.queue.submit(job_script,
                                               cwd=cwd)
                    self._jobs[job_id] = njob

            # Now check for finished jobs
            self._completed_jobs += [job_id for job_id in self._jobs
                                     if self.check_job(job_id,
                                                       **self._jobs[job_id])]

            for job_id in self._completed_jobs:
                cjob = self._jobs[job_id]
                # If needed, get the files from remote host
                if self.host is not None:
                    try:
                        self._getjob_remote(cjob)
                    except TimeoutError:
                        self.log('Timeout when trying to fetch results from'
                                 ' host. If it happens too frequently, try '
                                 'increasing the ssh_timeout argument.\n')
                        continue
                self.log('Job {0} completed\n'.format((cjob['name'])))
                self.finish_job(**cjob)
                # Remove the temporary directory
                shutil.rmtree(cjob['folder'])
                self.log('Folder {0} deleted\n'.format(cjob['folder']))
                # Finally delete it from our list
                del(self._jobs[job_id])

            # Only keep the ones that haven't been copied yet
            self._completed_jobs = list(set(self._completed_jobs) &
                                        set(self._jobs.keys()))

            sleep_time = self.check_time - (time.time()-loop_t0)
            sleep_time = sleep_time if sleep_time > 0 else 0
            time.sleep(sleep_time)

    def _putjob_remote(self, njob):
        """Copy the files generated for a job to a remote work directory"""
        with self.queue.remote_target.context as rtarg:
            rtarg.run_cmd('mkdir {0}'.format(njob['name']),
                          cwd=self.hostdir)
            rtarg.put_files(os.path.join(njob['folder'], '*'),
                            os.path.join(self.hostdir,
                                         njob['name']))

    def _getjob_remote(self, cjob):
        with self.queue.remote_target.context as rtarg:
            rpaths = [os.path.join(self.hostdir, cjob['name'],
                                   f.format(name=cjob['name'],
                                            **cjob['args']))
                      for f in self.remote_getfiles]
            rtarg.get_files(rpaths, cjob['folder'])
            rtarg.run_cmd('rm -rf {0}'.format(cjob['name']), cwd=self.hostdir)

    def _save(self):

        savedata = {}

        try:
            savedata.update(self.save_state())
        except TypeError:
            raise TypeError("Method save_state must return a dict")

        to_save = ['_jobs', '_waiting_jobs', '_completed_jobs']

        for k in to_save:
            savedata[k] = getattr(self, k)

        self.log('Saving data for future runs...\n')
        try:
            pickle.dump(savedata, open(self._pklname, 'w'))
        except IOError:
            self.log('Saving failed\n')

    def _load(self):

        to_load = ['_jobs', '_waiting_jobs', '_completed_jobs']

        self.log('Loading data from past runs...\n')
        try:
            loaddata = pickle.load(open(self._pklname))
        except IOError:
            self.log('Loading failed\n')
            return

        for k in to_load:
            setattr(self, k, loaddata[k])
            del loaddata[k]

        self.load_state(loaddata)

    def save_state(self):
        """Return a dictionary containing serialisable data to be saved from
        one run to the next"""
        return {}

    def load_state(self, loaded):
        """Replace attributes from loaded data in dictionary form"""
        return

    def next_job(self):
        """Return a dictionary definition of the next job in line"""
        return {'name': 'default_job', 'args': {}}

    def setup_job(self, name, args, folder):
        """Perform preparatory operations on the job"""
        return True

    def check_job(self, job_id, name, args, folder):
        """Checks if given job is complete or not"""
        return job_id not in self.queue.list()

    def finish_job(self, name, args, folder):
        """Performs completion operations on the job. At this point any
        relevant output files should be copied from 'folder' to their final
        destination as the temporary folder itself will be deleted immediately
        after"""
        pass

    def start_run(self):
        """Operations to perform when the daemon thread starts running"""
        pass

    def finish_run(self):
        """Operations to perform after the daemon thread stops running"""
        pass

    def log(self, logtxt):
        self._log.write(logtxt)
        self._log.flush()

    @staticmethod
    def stop(fname, subname):
        """Stop Submitter process from filename and name,
        return False if failed"""
        return sp.Popen(['pkill', '-f', fname + ' ' + subname]).wait() == 0

    @staticmethod
    def list():
        list_proc = sp.Popen(['ps', 'aux'],
                             stdin=sp.PIPE,
                             stdout=sp.PIPE,
                             stderr=sp.PIPE)
        stdout, stderr = safe_communicate(list_proc)
        # Parse stdout
        all_subms = []
        for l in stdout.split('\n'):
            if 'soprano.hpc.submitter._spawn' in l:
                lspl = l.split()
                # File, name, time, PID
                subm = lspl[-3], lspl[-2], lspl[-7], int(lspl[1])
                all_subms.append(subm)

        return all_subms

    @property
    def _pklname(self):
        return self.name + '_save.pkl'



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generator producing molecular neighbourhoods"""

import numpy as np
from ase import Atoms
# Internal imports
from soprano.utils import minimum_supcell, supcell_gridgen


def molecularNeighbourhoodGen(struct, mols, central_mol=0, max_R=10,
                              method='com', use_supercell=False):
    """Generator function to create a spherical molecular neighbourhood. Given
    a structure and its molecules as returned by the Molecules property, 
    produce supercell structures that contain one molecule each, progressively
    further away from the one indicated as central.

    | Args:
    |   struct (ase.Atoms): original structure
    |   mols ([ase.AtomsSelection]): list of molecules, as returned by the
    |                                soprano.properties.linkage.Molecules 
    |                                class.
    |   central_mol (int): index of the molecule whose centre of mass is 
    |                      considered central. Default is 0.
    |   max_R (float): maximum radius of the neighbourhood sphere. Default is 
    |                  10 Ang.
    |   method (str): method to compute distance between molecules. 'com' 
    |                 means using the center of mass. 'nearest' means using
    |                 the closest atom. Default is 'com'.
    |   use_supercell (bool): if True, all returned structures will have a 
    |                         cell large enough to contain the entire
    |                         neighbourhood. Default is False.

    | Returns:
    |   molecularNeighbourhoodGen (generator): an iterator object that yields
    |                                         structures within the given 
    |                                         spherical neighbourhood.                                         

    """

    # Supercell size?
    scell = minimum_supcell(max_R, struct.get_cell())
    fgrid, grid = supcell_gridgen(struct.get_cell(), scell)

    # Center?
    mol_structs = [m.subset(struct) for m in mols]
    mol_coms = np.array([a.get_center_of_mass() for a in mol_structs])

    # Origin
    p0 = mol_coms[central_mol]

    # Positions?
    if method == 'com':
        positions = mol_coms[None, :, :]+grid[:, None, :]-p0
    elif method == 'nearest':
        positions = np.zeros((len(grid), len(mols), 3))
        for i, a in enumerate(mol_structs):
            dp = a.get_positions() - p0
            p = dp[None, :, :] + grid[:, None, :]
            # Closest one?
            positions[:, i, :] = p[range(len(grid)),
                                   np.argmin(np.linalg.norm(p, axis=-1),
                                             axis=1)]
    else:
        raise RuntimeError('Invalid method passed to '
                           'molecularNeighbourhoodGen')

    # Order of appearance?
    distances = np.linalg.norm(positions, axis=-1)
    sphere = np.where(distances <= max_R)
    distances = distances[sphere[0], sphere[1]]
    fxyz = fgrid[sphere[0]]
    xyz = grid[sphere[0]]
    mol_i = np.arange(len(mols))[sphere[1]]
    positions = positions[sphere[0], sphere[1]]

    # Now the order
    order = np.argsort(distances)

    for i in order:
        a = mol_structs[mol_i[i]].copy()
        # Create the structure
        if use_supercell:
            a.set_cell(np.dot(np.diag(scell), a.get_cell()))
        a.set_positions(a.get_positions() + xyz[i])

        # Add some info
        a.info['neighbourhood_info'] = {
            'molecule_index': mol_i[i],
            'molecule_cell': fxyz[i],
            'molecule_distance': distances[i]
        }

        yield a



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Module containing AtomsProperties that relate to NMR properties of a system.
Some of these are valid only for Atoms objects loaded from a .magres file.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import

from soprano.properties.nmr.ms import (MSDiagonal, MSIsotropy, MSAnisotropy,
                                       MSReducedAnisotropy, MSAsymmetry,
                                       MSSpan, MSSkew, MSQuaternion)
from soprano.properties.nmr.efg import (EFGDiagonal, EFGVzz, EFGAnisotropy,
                                        EFGReducedAnisotropy, EFGAsymmetry,
                                        EFGSpan, EFGSkew,
                                        EFGQuadrupolarConstant,
                                        EFGQuaternion)
from soprano.properties.nmr.dipolar import (DipolarCoupling, DipolarDiagonal,
                                            DipolarRSS, DipolarTensor)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generator producing structures by repeatedly applying a transform"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import numpy as np
from ase import Atoms
# Internal imports
import soprano.utils as utils
from soprano.properties import AtomsProperty


def transformGen(struct_0, transform, steps=10):
    """Generator function to create multiple structures with positions
    interpolated linearly between two extremes.

    | Args:
    |   struct_0 (ase.Atoms): the starting structure
    |   transform (function): the transform to apply on the given structure.
    |                         Must accept and return a structure as only
    |                         arguments. Instances of all transforms from the
    |                         soprano.properties.transform module satisfy
    |                         these conditions.
    |   steps (Optional[int]): number of times to apply the transform.
    |                          Default is 10

    | Returns:
    |   linspaceGenerator (generator): an iterator object that yields
    |                                  structures created by linear
    |                                  interpolation.

    """

    # Sanity check
    if not hasattr(transform, '__call__'):
        raise ValueError('Invalid transform function passed to transformGen')

    rootname = struct_0.info['name'] if 'name' in struct_0.info else 'transf'
    struct = struct_0

    for i in range(steps):

        struct = transform(struct)
        if type(struct) is not Atoms:
            raise RuntimeError('Invalid return value from transform')
        struct.info['name'] = '{0}_{1}'.format(rootname, i)

        yield struct



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of QueueInterface class.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import subprocess as sp

from soprano.utils import is_string, safe_communicate
from soprano.hpc.submitter.utils import RemoteTarget


class QueueInterface(object):

    """QueueInterface object

    A class meant to simplify interfacing in a basic way
    with a Queue system. Contains commands to submit to the queue, list the
    job IDs, and kill them if necessary. Will contain Regexps to parse for IDs
    and additional information as returned upon submission and listing.
    It is important that the regular expressions used employ NAMED GROUPS to
    parse the various fields. In particular, a job_id group must ALWAYS be
    present.
    The class also provides some static variables implementing standard
    interfaces for common queueing systems. These can be retrieved by using
    QueueInterface.<NAME>. The currently implemented names are the following:

    - LSF (IBM's managing system, using the command bsub)
    - GridEngine (Sun's managing system, also available in an open version,
                  using the command qsub)
    - PBS (another managing system using qsub)
    """

    def __init__(self, sub_cmd, list_cmd, kill_cmd, sub_outre, list_outre,
                 list_user_opt=None):
        """Initialize the QueueInterface.

        | Args:
        |   sub_cmd (str): command used to submit a script to the queue
        |   list_cmd (str): command used to list all queued jobs for the user
        |   kill_cmd (str): command used to kill a job given its id
        |   sub_outre (str): regular expression used to parse the output of
        |                    sub_cmd. Must contain at least a job_id named
        |                    group
        |   list_outre (str): regular expression used to parse the output of
        |                     list_cmd. Must contain at least a job_id named
        |                     group
        |   list_user_opt (Optional[str]): name of the option for passing a
        |                                  specific user name when listing
        |                                  jobs. For example, on LSF this is
        |                                  -u. By default it's None, and the
        |                                  default of the queueing system is
        |                                  used.

        """

        self.sub_cmd = sub_cmd
        self.list_cmd = list_cmd
        self.kill_cmd = kill_cmd

        self.sub_outre = re.compile(sub_outre)
        if 'job_id' not in self.sub_outre.groupindex:
            raise ValueError('sub_outre does not contain job_id group')
        self.list_outre = re.compile(list_outre)
        if 'job_id' not in self.list_outre.groupindex:
            raise ValueError('list_outre does not contain job_id group')

        self.list_user_opt = list_user_opt

        self._rTarg = None

    def set_remote_host(self, host=None, timeout=1.0):
        """Set a remote host for use by this QueueInterface

        If a remote host is set, this Interface will try using the Paramiko
        library to connect to it via SSH. Check the docstring of
        soprano.hpc.submitter.utils.RemoteTarget to see what the limitations
        are for this usage.

        | Args:
        |   host (Optional[str]): host name of the remote machine to connect
        |                         to. Further information must be contained in
        |                         the user's ~/.ssh/config file. If left empty
        |                         the remote host is UNSET and therefore the
        |                         interface goes back to local use.
        |   timeout (Optional[float]): connection timeout in seconds (default
        |                              is 1 second)

        """

        if host is not None:
            self._rTarg = RemoteTarget(host, timeout)
        else:
            self._rTarg = None

    def submit(self, script, cwd=None):
        """Submit a job to the queue.x

        | Args:
        |   script (str): content of the submission script
        |   cwd (Optional[str]): path to the desired working directory
        |
        | Returns:
        |   job_id (str): the job ID assigned by the queue system and parsed
        |                 with sub_outre
        """

        if self._rTarg is None:
            subproc = sp.Popen(self.sub_cmd.split(), stdin=sp.PIPE,
                               stdout=sp.PIPE,
                               stderr=sp.PIPE,
                               cwd=cwd)

            stdout, stderr = safe_communicate(subproc, script)
        else:
            with self._rTarg.context as rTarg:
                stdout, stderr = rTarg.run_cmd(self.sub_cmd,
                                               cwd=cwd, stdin=script)

        # Parse out the job id!
        match = self.sub_outre.search(stdout)
        if match is None:
            raise RuntimeError('Submission of job has failed with output:\n'
                               '\tSTDOUT: {0}\n\tSTDERR: {1}'.format(stdout,
                                                                     stderr))
        else:
            return match.groupdict()['job_id']

    def list(self, user='$USER'):
        """List all jobs found in the queue

        | Returns:
        |   jobs (dict): a dict of jobs classified by ID containing all info
        |                that can be matched through list_outre
        |   user (Optional[str]): user for whom jobs should be listed. Will
        |                         not have any effect if list_user_opt has not
        |                         been specified. Default is $USER.
        |

        """

        cmd = self.list_cmd
        if self.list_user_opt is not None:
            cmd += ' {0} {1}'.format(self.list_user_opt, user)

        if self._rTarg is None:
            subproc = sp.Popen(cmd.split(), stdout=sp.PIPE,
                               stderr=sp.PIPE)

            stdout, stderr = safe_communicate(subproc)
        else:
            with self._rTarg.context as rTarg:
                stdout, stderr = rTarg.run_cmd(cmd)

        # Parse out everything!
        jobs = {}
        for line in stdout.split('\n'):
            match = self.list_outre.search(line)
            if match is None:
                continue
            else:
                jobdict = match.groupdict()
                jobs[jobdict['job_id']] = jobdict

        return jobs

    def kill(self, job_id):
        """Kill the job with the given ID

        | Args:
        |   job_id (str): ID of the job to kill
        |
        """

        if self._rTarg is None:
            subproc = sp.Popen(self.kill_cmd.split() + [job_id],
                               stdout=sp.PIPE,
                               stderr=sp.PIPE)
            stdout, stderr = safe_communicate(subproc)
        else:
            with self._rTarg.context as rTarg:
                stdout, stderr = rTarg.run_cmd(self.kill_cmd +
                                               ' {0}'.format(job_id))

    @property
    def remote_target(self):
        """RemoteTarget if a remote Host is set"""
        return self._rTarg

    @classmethod
    def LSF(cls):
        return cls(sub_cmd='bsub',
                   list_cmd='bjobs',
                   kill_cmd='bkill',
                   sub_outre='Job \<(?P<job_id>[0-9]+)\>',
                   list_outre='(?P<job_id>[0-9]+)[^(RUN|PEND)]*'
                              '(?P<job_status>RUN|PEND)',
                   list_user_opt='-u')

    @classmethod
    def GridEngine(cls):
        return cls(sub_cmd='qsub',
                   list_cmd='qstat',
                   kill_cmd='qdel',
                   sub_outre='Your job (?P<job_id>[0-9]+)',
                   list_outre='(?P<job_id>[0-9]+)\s.*'
                              '\s(?P<job_status>r|qw)\s',
                   list_user_opt='-u')

    @classmethod
    def PBS(cls):
        return cls(sub_cmd='qsub',
                   list_cmd='qstat',
                   kill_cmd='qdel',
                   sub_outre='(?P<job_id>[^\s]+)',
                   list_outre='(?P<job_id>[^\s]+)\s.*'
                              '\s(?P<job_status>R|Q)\s',
                   list_user_opt='-u')



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
NMR Data

Data on NMR relevant properties of elements and isotopes - spin, gyromagnetic
ratio and quadrupole moment.
"""

import re
import json
import pkgutil
import numpy as np
import scipy.constants as cnst

# EFG conversion constant.
# Units chosen so that EFG_TO_CHI*Quadrupolar moment*Vzz = Hz
EFG_TO_CHI = cnst.physical_constants['atomic unit of electric field '
                                     'gradient'][0]*cnst.e*1e-31/cnst.h

try:
    _nmr_data = pkgutil.get_data('soprano',
                                 'data/nmrdata.json').decode('utf-8')
    _nmr_data = json.loads(_nmr_data)
except IOError:
    _nmr_data = None


def _get_nmr_data():

    if _nmr_data is not None:
        return _nmr_data
    else:
        raise RuntimeError('NMR data not available. Something may be '
                           'wrong with this installation of Soprano')


def _get_isotope_data(elems, key, isotopes={}, isotope_list=None,
                      use_q_isotopes=False):

    data = np.zeros(len(elems))
    nmr_data = _get_nmr_data()

    for i, e in enumerate(elems):

        if e not in nmr_data:
            # Non-existing element
            raise RuntimeError('No NMR data on element {0}'.format(e))

        iso = nmr_data[e]['iso']
        if use_q_isotopes and nmr_data[e]['Q_iso'] is not None:
            iso = nmr_data[e]['Q_iso']
        if e in isotopes:
            iso = isotopes[e]
        if isotope_list is not None and isotope_list[i] is not None:
            iso = isotope_list[i]

        try:
            data[i] = nmr_data[e][str(iso)][key]
        except KeyError:
            raise RuntimeError('Data {0} does not exist for isotope {1} of '
                               'element {2}'.format(key, iso, e))

    return data


def _el_iso(sym):
    """ Utility function: split isotope and element in conventional
    representation.
    """

    nmr_data = _get_nmr_data()

    match = re.findall('([0-9]*)([A-Za-z]+)', sym)
    if len(match) != 1:
        raise ValueError('Invalid isotope symbol')
    elif match[0][1] not in nmr_data:
        raise ValueError('Invalid element symbol')

    el = match[0][1]
    # What about the isotope?
    iso = str(nmr_data[el]['iso']) if match[0][0] == '' else match[0][0]

    if iso not in nmr_data[el]:
        raise ValueError('No data on isotope {0} for element {1}'.format(iso,
                                                                         el))

    return el, iso


def nmr_gamma(el, iso=None):
    """Gyromagnetic ratio for an element

    Return the gyromagnetic ratio for the given element and isotope, in
    rad/(s*T)

    | Args:
    |   el (str):   element symbol
    |   iso (int):  isotope. Default is the most abundant one.

    | Returns:
    |   gamma (float):  gyromagnetic ratio in rad/(s*T)    
    """

    isotopes = {}
    if iso is not None:
        isotopes[el] = iso

    return _get_isotope_data([el], 'gamma', isotopes=isotopes)


def nmr_spin(el, iso=None):
    """Nuclear spin for an element

    Return the nuclear spin for the given element and isotope, in
    Bohr magnetons

    | Args:
    |   el (str):   element symbol
    |   iso (int):  isotope. Default is the most abundant one.

    | Returns:
    |   I (float):  nuclear spin in Bohr magnetons
    """

    isotopes = {}
    if iso is not None:
        isotopes[el] = iso

    return _get_isotope_data([el], 'I', isotopes=isotopes)


def nmr_quadrupole(el, iso=None):
    """Quadrupole moment for an element

    Return the quadrupole moment for the given element and isotope, in
    barns

    | Args:
    |   el (str):   element symbol
    |   iso (int):  isotope. Default is the most abundant one.

    | Returns:
    |   Q (float):  quadrupole moment in barns
    """

    isotopes = {}
    if iso is not None:
        isotopes[el] = iso

    return _get_isotope_data([el], 'Q', isotopes=isotopes)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Contains the NMRTensor class, simplifying the process of diagonalisation of an
NMR tensor as well as its representation in multiple conventions
"""

import numpy as np
from soprano.nmr.utils import (_evals_sort, _haeb_sort, _anisotropy,
                               _asymmetry, _span, _skew, _evecs_2_quat,
                               _dip_constant)
from soprano.data.nmr import _get_isotope_data
from ase.quaternions import Quaternion


class NMRTensor(object):
    """NMRTensor

    Class containing an NMR tensor, useful to access all its most important
    properties and representations.
    """

    ORDER_INCREASING = 'i'
    ORDER_DECREASING = 'd'
    ORDER_HAEBERLEN = 'h'
    ORDER_NQR = 'n'

    def __init__(self, data, order=ORDER_INCREASING):
        """
        Initialise the NMRTensor

        Create an NMRTensor object from a 3x3 matrix.

        Arguments:
            data (np.ndarray or tuple):  3x3 matrix containing the tensor, or
                                         pair [evals, evecs] for the symmetric
                                         part alone.
            order (str):        Order to use for eigenvalues/eigenvectors. Can
                                be 'i' (ORDER_INCREASING), 'd' 
                                (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                                'n' (ORDER_NQR). Default is 'i'.
        """

        self._order = order

        if len(data) == 3:
            self._data = np.array(data)
            if self._data.shape != (3, 3):
                raise ValueError('Invalid matrix data passed to NMRTensor')
            self._symm = (self._data+self._data.T)/2.0
            evals, evecs = np.linalg.eigh(self._symm)
        elif len(data) == 2:
            evals, evecs = data
            evecs = np.array(evecs)
            sort_i = np.argsort(evals)
            if len(evals) != 3 or evecs.shape != (3, 3):
                raise ValueError('Invalid eigenvalues/vectors passed to '
                                 'NMRTensor')
            evals = evals[sort_i]
            evecs = evecs[:, sort_i]
            self._symm = np.linalg.multi_dot([evecs, np.diag(evals), evecs.T])
            self._data = self._symm

        self._incr_evals = evals
        self._haeb_evals = _haeb_sort([evals])[0]
        self._anisotropy = None
        self._redaniso = None
        self._asymmetry = None
        self._span = None
        self._skew = None

        self._trace = None

        # Spherical tensor components
        self._sph = None

        # Sort eigenvalues and eigenvectors as specified
        if order != self.ORDER_INCREASING:
            self._evals, sort_i = _evals_sort([evals], order, True)
            self._evals = self._evals[0]
            self._evecs = evecs[:, sort_i[0]]
        else:
            # No point in fixing what ain't broken
            self._evals = evals
            self._evecs = evecs

        # Last eigenvector must be the cross product of the first two
        # (apparently this is much faster than np.cross. Beats me why)
        self._evecs[0, 2] = (self._evecs[1, 0]*self._evecs[2, 1] -
                             self._evecs[2, 0]*self._evecs[1, 1])
        self._evecs[1, 2] = (self._evecs[2, 0]*self._evecs[0, 1] -
                             self._evecs[0, 0]*self._evecs[2, 1])
        self._evecs[2, 2] = (self._evecs[0, 0]*self._evecs[1, 1] -
                             self._evecs[1, 0]*self._evecs[0, 1])

        self._quat = None

    @property
    def data(self):
        return self._data.copy()

    @property
    def eigenvalues(self):
        return self._evals.copy()

    @property
    def eigenvectors(self):
        return self._evecs.copy()

    @property
    def trace(self):
        if self._trace is None:
            self._trace = np.trace(self._data)
        return self._trace

    @property
    def isotropy(self):
        return self.trace/3.0

    @property
    def anisotropy(self):
        if self._anisotropy is None:
            self._anisotropy = _anisotropy(self._haeb_evals[None, :])[0]
        return self._anisotropy

    @property
    def reduced_anisotropy(self):
        if self._redaniso is None:
            self._redaniso = _anisotropy(self._haeb_evals[None, :], True)[0]
        return self._redaniso

    @property
    def asymmetry(self):
        if self._asymmetry is None:
            self._asymmetry = _asymmetry(self._haeb_evals[None, :])[0]
        return self._asymmetry

    @property
    def span(self):
        if self._span is None:
            self._span = _span(self._incr_evals[None, :])[0]
        return self._span

    @property
    def skew(self):
        if self._skew is None:
            self._skew = _skew(self._incr_evals[None, :])[0]
        return self._skew

    @property
    def quaternion(self):
        if self._quat is None:
            self._quat = _evecs_2_quat([self._evecs])[0]
        return Quaternion(self._quat.q)

    @property
    def spherical_repr(self):
        if self._sph is None:
            self._sph = np.zeros((3, 3, 3))
            self._sph[0] = np.eye(3)*self.trace/3
            self._sph[1] = (self._data-self._data.T)/2.0
            self._sph[2] = self._symm - self._sph0

        return self._sph.copy()

    def euler_angles(self, convention='zyz'):
        """Return Euler angles of the Principal Axis System

        Return Euler angles of the PAS for this tensor in the
        required convention (currently supported: zyz, zxz).

        Keyword Arguments:
            convention {str} -- Euler angles convention to use
            (default: {'zyz'})
        """

        convention = convention.lower()
        return self.quaternion.euler_angles(convention)

    @staticmethod
    def make_dipolar(a, i, j, cell=[0, 0, 0], isotopes={}, isotope_i=None,
                     isotope_j=None, rotation_axis=None):
        """Create a dipolar NMRTensor

        Create a dipolar NMR tensor from an atoms object and the indices
        of two atoms. Values are in Hz.

        | Args:
        |   a (ase.Atoms):      Atoms object of the structure to compute the 
        |                       tensor for
        |   i (int):            index of first atom
        |   j (int):            index of second atom
        |   cell (np.array):    vector of the cell of the second atom, for 
        |                       couplings between atoms in different cells.
        |                       By default is [0,0,0].
        |   isotopes (dict): dictionary of specific isotopes to use, by element
        |                    symbol. If the isotope doesn't exist an error will
        |                    be raised.
        |   isotope_i (int): isotope of atom i. To be used if
        |                    different atoms of the same element are supposed
        |                    to be of different isotopes. If None it will fall
        |                    back on the previous definitions. Otherwise it
        |                    overrides everything else.
        |   isotope_j (int): isotope of atom j. See above.
        |   rotation_axis (np.array):   an axis around which the selected pair
        |                               is rotating. If present, the tensor
        |                               will be averaged for infinitely fast
        |                               rotation around it.

        | Returns:
        |   diptens (NMRTensor):    an NMRTensor object with the dipolar 
        |                           coupling matrix as data.

        """

        pos = a.get_positions()
        elems = np.array(a.get_chemical_symbols())
        r = pos[j]-pos[i] + np.dot(a.get_cell(), cell)
        rnorm = np.linalg.norm(r)
        gammas = _get_isotope_data(elems[[i, j]], 'gamma', isotopes=isotopes,
                                   isotope_list=[isotope_i, isotope_j])

        d = _dip_constant(rnorm*1e-10, *gammas)

        evals = np.zeros(3)
        evecs = np.zeros((3, 3))

        if rotation_axis is None:
            axis = r/rnorm
        else:
            axis = np.array(rotation_axis)/np.linalg.norm(rotation_axis)
            vp2 = np.dot(r/rnorm, axis)**2
            d = 0.5*d*(3*vp2-1)

        evals[2] = d
        evals[:2] = -d/2.0

        evecs[:, 2] = axis

        x, y = axis[:2]
        evecs[1, 0] = x
        evecs[0, 0] = -y
        evecs[:, 0] /= (x**2+y**2)**0.5

        evecs[0, 1]
        evecs[0, 1] = -evecs[2, 2]*evecs[1, 0]
        evecs[1, 1] = evecs[2, 2]*evecs[0, 0]
        evecs[2, 1] = evecs[0, 2]*evecs[1, 0]-evecs[1, 2]*evecs[0, 0]

        if d < 0:
            evals = evals[::-1]
            evecs = evecs[:, ::-1]

        D = [evals, evecs]

        return NMRTensor(D)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of some CASTEP related AtomsProperties"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import os
from soprano.properties import AtomsProperty


class CastepEnthalpy(AtomsProperty):

    """
    CastepEnthalpy

    Enthalpy as found in the .castep file of a GeometryOptimization
    calculation. If not present, this will fall back on the final free energy.

    | Parameters:
    |   castep_path (str): the path in which the .castep file is to be found.
    |   seedname_info (str): the Atoms.info key that contains the seedname
    |                        of the .castep file. By default is 'name'.

    """

    default_name = 'castep_enthalpy'
    default_params = {
        'castep_path': '.',
        'seedname_info': 'name'
    }

    @staticmethod
    def extract(s, castep_path, seedname_info):
        # Open the file
        fname = os.path.join(castep_path, s.info[seedname_info] + '.castep')
        f = open(fname).read()
        # Parse the expression of interest
        ent_re = re.compile('BFGS: Final Enthalpy\s+=\s+([\+\-\.0-9E]+)\s+eV')
        match = ent_re.findall(f)
        if len(match) > 0:
            try:
                return float(match[-1])
            except:
                raise RuntimeError('Can\'t parse enthalpy:'
                                   'corrupted CASTEP file')
        else:
            # Settle for the final free energy
            nrg_re = re.compile('Final free energy \(E\-TS\)\s+=\s+'
                                '([\+\-\.0-9E]+)\s+eV')
            match = nrg_re.findall(f)
            if len(match) == 0:
                raise RuntimeError('Corrupted or incomplete CASTEP file')
            return float(match[-1])



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to NMR dipole-dipole
couplings"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy import constants as cnst
from soprano.utils import minimum_periodic, minimum_supcell, supcell_gridgen
from soprano.properties import AtomsProperty
from soprano.selection import AtomSelection
from soprano.nmr.utils import _dip_constant
from soprano.data.nmr import _get_isotope_data


class DipolarCoupling(AtomsProperty):

    """
    DipolarCoupling

    Produces a dictionary of dipole-dipole coupling constants for atom pairs
    in the system. For each pair, the closest periodic copy will be considered.
    The constant for a pair of nuclei i and j is defined as:

    .. math::

        d_{ij} = -\\frac{\\mu_0\\hbar\\gamma_i\\gamma_j}{8\\pi^2r_{ij}^3}

    where the gammas represent the gyromagnetic ratios of the nuclei and the
    r is their distance. The full tensor of the interaction is then defined as

    .. math::

         D_{ij} = 
         \\begin{bmatrix}
          -\\frac{d_{ij}}{2} & 0 & 0 \\\\
          0 & -\\frac{d_{ij}}{2} & 0 \\\\
          0 & 0 & d_{ij} \\\\
         \\end{bmatrix}

    where the z-axis is aligned with :math:`r_{ij}` and the other two can be any
    directions in the orthogonal plane.

    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to compute the dipolar
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to compute the dipolar
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with its
    |                         own closest periodic copy. Otherwise excluded.
    |                         Default is False.
    |   block_size (int): maximum size of blocks used when processing large
    |                     chunks of pairs. Necessary to avoid memory problems
    |                     for very large systems. Default is 1000.

    | Returns: 
    |   dip_dict (dict): Dictionary of couplings in Hz and r_{ij} versors,
    |                    pointing from i to j, by atomic index pair.

    """

    default_name = 'dip_coupling'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'block_size': 1000,
    }

    @staticmethod
    def extract(s, sel_i, sel_j, isotopes, isotope_list, self_coupling,
                block_size):

        # Selections
        if sel_i is None:
            sel_i = AtomSelection.all(s)
        elif not isinstance(sel_i, AtomSelection):
            sel_i = AtomSelection(s, sel_i)

        if sel_j is None:
            sel_j = sel_i
        elif not isinstance(sel_j, AtomSelection):
            sel_j = AtomSelection(s, sel_j)

        # Find gammas
        elems = s.get_chemical_symbols()

        gammas = _get_isotope_data(elems, 'gamma', isotopes, isotope_list)

        # Viable pairs
        pairs = [(i, j) for i in sel_i.indices
                 for j in sel_j.indices]
        if not self_coupling:
            pairs = [p for p in pairs if p[0] != p[1]]

        pairs = np.array(pairs).T
        # Need to sort them and remove any duplicates, also take i < j as
        # convention
        pairs = np.array(list(zip(*set([tuple(x)
                                        for x in np.sort(pairs, axis=0).T]))))

        pos = s.get_positions()

        # Split this in blocks to make sure we don't clog the memory

        d_ij = np.zeros((0,))
        v_ij = np.zeros((0, 3))

        npairs = pairs.shape[1]

        for b_i in range(0, npairs, block_size):
            block = pairs.T[b_i:b_i+block_size]
            r_ij = pos[block[:, 1]] - pos[block[:, 0]]
            # Reduce to NN
            r_ij, _ = minimum_periodic(r_ij, s.get_cell(), exclude_self=True)
            # Distance
            R_ij = np.linalg.norm(r_ij, axis=1)
            # Versors
            v_ij = np.concatenate([v_ij, r_ij/R_ij[:, None]], axis=0)
            # Couplings
            d_ij = np.concatenate([d_ij,
                                   _dip_constant(R_ij*1e-10,
                                                 gammas[block[:, 0]],
                                                 gammas[block[:, 1]])
                                   ])

        return {tuple(ij): [d_ij[l], v_ij[l]] for l, ij in enumerate(pairs.T)}


class DipolarTensor(AtomsProperty):

    """
    DipolarTensor

    Produces a dictionary of dipole-dipole coupling tensors for atom pairs
    in the system. For each pair, the closest periodic copy will be considered.
    The coupling constant for a pair of nuclei i and j is defined as:

    .. math::

        d_{ij} = -\\frac{\\mu_0\\hbar\\gamma_i\\gamma_j}{8\\pi^2r_{ij}^3}

    where the gammas represent the gyromagnetic ratios of the nuclei and the
    r is their distance. The full tensor of the interaction is then defined as

    .. math::

         D_{ij} = d_{ij}\\frac{3\\hat{r}_{ij}\\otimes \\hat{r}_{ij}-\\mathbb{I}}{2}

    where :math:`\\hat{r}_{ij} = r_{ij}/|r_{ij}|` and the Kronecker product is
    used.

    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to compute the dipolar
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to compute the dipolar
    |                                   coupling with the ones in sel_i. By
    |                                   default is None (= same as sel_i).
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with its
    |                         own closest periodic copy. Otherwise excluded.
    |                         Default is False.
    |   block_size (int): maximum size of blocks used when processing large
    |                     chunks of pairs. Necessary to avoid memory problems
    |                     for very large systems. Default is 1000.

    | Returns: 
    |   dip_dict (dict): Dictionary of tensors in Hz by atomic index pair.

    """

    default_name = 'dip_coupling'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'block_size': 1000,
    }

    @staticmethod
    def extract(s, sel_i, sel_j, isotopes, isotope_list, self_coupling,
                block_size):

        dip_dict = DipolarCoupling.extract(s,
                                           sel_i=sel_i, sel_j=sel_j,
                                           isotopes=isotopes,
                                           isotope_list=isotope_list,
                                           self_coupling=self_coupling,
                                           block_size=block_size)

        # Now build the tensors
        tdict = {}
        for ij, (d, r) in dip_dict.items():
            tdict[ij] = d*(3*r[:, None]*r[None, :]-np.eye(3))/2

        return tdict


class DipolarDiagonal(AtomsProperty):

    """
    DipolarDiagonal

    Produces a dictionary of dipole-dipole tensors as eigenvalues and
    eigenvectors for atom pairs in the system. For each pair, the closest
    periodic copy will be considered.

    | Parameters:
    |   sel_i (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to compute the dipolar
    |                                   coupling. By default is None
    |                                   (= all of them).
    |   sel_j (AtomSelection or [int]): Selection or list of indices of atoms
    |                                   for which to compute the dipolar
    |                                   coupling with the ones i sel_i. By
    |                                   default is None (= same as sel_i).
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.
    |   self_coupling (bool): if True, include coupling of a nucleus with its
    |                         own closest periodic copy. Otherwise excluded.
    |                         Default is False.
    |   block_size (int): maximum size of blocks used when processing large
    |                     chunks of pairs. Necessary to avoid memory problems
    |                     for very large systems. Default is 1000.

    | Returns: 
    |   dip_tens_dict (dict): Dictionary of dipolar eigenvalues (in Hz) and
    |                         eigenvectors, by atomic index pair.

    """

    default_name = 'dip_diagonal'
    default_params = {
        'sel_i': None,
        'sel_j': None,
        'isotopes': {},
        'isotope_list': None,
        'self_coupling': False,
        'block_size': 1000
    }

    @staticmethod
    def extract(s, sel_i, sel_j, isotopes, isotope_list, self_coupling,
                block_size):

        # First, just get the values
        dip_dict = DipolarCoupling.extract(s, sel_i, sel_j,
                                           isotopes, isotope_list,
                                           self_coupling, block_size)

        # Now build the tensors
        dip_tens_dict = {}

        for ij, (d, v) in dip_dict.items():

            evals = np.array([-d/2, -d/2, d])
            # Eigenvectors
            evecs = np.zeros((3, 3))
            # Z is equal to v
            evecs[:, 2] = v
            # Y is any random orthogonal vector
            rv = np.random.random(3)
            evecs[:, 1] = np.cross(v, rv)
            evecs[:, 1] /= np.linalg.norm(evecs[:, 1])
            # X = Y cross Z
            evecs[:, 0] = np.cross(evecs[:, 1], v)

            dip_tens_dict[ij] = {'evals': evals, 'evecs': evecs}

        return dip_tens_dict


class DipolarRSS(AtomsProperty):

    """
    DipolarRSS

    Compute the Dipolar constant Root Sum Square for each atom in a system,
    including periodicity, within a cutoff.

    | Parameters:
    |   cutoff (float): cutoff radius in Angstroms at which the sum stops. By
    |                   default 5 Ang.
    |   isonuclear (bool): if True, only nuclei of the same species will be
    |                      considered. By default is False.
    |   isotopes (dict): dictionary of specific isotopes to use, by element
    |                    symbol. If the isotope doesn't exist an error will
    |                    be raised.
    |   isotope_list (list): list of isotopes, atom-by-atom. To be used if
    |                        different atoms of the same element are supposed
    |                        to be of different isotopes. Where a 'None' is
    |                        present will fall back on the previous
    |                        definitions. Where an isotope is present it
    |                        overrides everything else.

    | Returns:
    |   dip_rss (np.ndarray): dipolar constant RSS for each atom in the system

    """

    default_name = 'dip_rss'
    default_params = {
        'cutoff': 5.0,
        'isonuclear': False,
        'isotopes': {},
        'isotope_list': None,
    }

    @staticmethod
    def extract(s, cutoff, isonuclear, isotopes, isotope_list):

        # Supercell size
        scell_shape = minimum_supcell(cutoff, s.get_cell())
        _, scell = supcell_gridgen(s.get_cell(), scell_shape)

        pos = s.get_positions()
        elems = np.array(s.get_chemical_symbols())

        gammas = _get_isotope_data(elems, 'gamma', isotopes, isotope_list)

        dip_rss = []

        for i, el in enumerate(elems):

            # Distances?
            if not isonuclear:
                rij = pos.copy()
                gj = np.tile(gammas, len(scell))
            else:
                rij = pos[np.where(elems == el)]
                gj = gammas[i]
            rij = rij[None, :, :]+scell[:, None, :]-pos[i, None, None]
            Rij = np.linalg.norm(rij.reshape((-1, 3)), axis=-1)
            # Valid indices?
            ij = np.where((Rij > 0) & (Rij <= cutoff))
            Rij = Rij[ij]*1e-10
            try:
                gj = gj[ij]
            except IndexError:
                pass

            dip = _dip_constant(Rij, gammas[i], gj)
            dip_rss.append(np.sqrt(np.sum(dip**2)))

        return np.array(dip_rss)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomProperties that relate to symmetry"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from collections import namedtuple
from soprano.properties import AtomsProperty
from soprano.properties.symmetry.utils import (_get_symmetry_dataset,
                                               _find_wyckoff_points)

WyckoffPoint = namedtuple('WyckoffPoint', ['fpos', 'pos', 'operations',
                                           'hessian'])


class SymmetryDataset(AtomsProperty):

    """
    SymmetryDataset

    Extracts SPGLIB's standard symmetry dataset from a given system, including
    spacegroup symbol, symmetry operations etc.

    | Parameters:
    |   symprec (float): distance tolerance, in Angstroms, applied when
    |                    searching symmetry.

    | Returns:
    |   symm_dataset (dict): dictionary of symmetry information

    """

    default_name = 'symmetry_dataset'
    default_params = {
        'symprec': 1e-5,
    }

    @staticmethod
    def extract(s, symprec):
        return _get_symmetry_dataset(s, symprec=symprec)


class WyckoffPoints(AtomsProperty):

    """
    WyckoffPoints

    Returns a list of the found high symmetry points for a given system,
    including information about their point group operations, and the 
    properties of Hessian-like quantities at that point, namely, if they are
    constrained to be isotropic, definite (positive/negative), or
    can be anything.


    | Parameters:
    |   symprec (float): distance tolerance, in Angstroms, applied when
    |                    searching symmetry.

    | Returns:
    |   wyckoff_points (list): a list of WyckoffPoint named tuples, containing
    |                          the members 'fpos' (fractional coordinates),
    |                          'pos' (Cartesian coordinates), 'operations'
    |                          (point group operations) and 'isotropic'
    |                          (whether Hessian-like tensors are isotropic at
    |                          the point).

    """

    default_name = 'wyckoff_points'
    default_params = {
        'symprec': 1e-5,
    }

    @staticmethod
    def extract(s, symprec):

        hprops = ['saddle', 'none', 'definite', 'isotropic']

        fpos, ops, hess = _find_wyckoff_points(s, symprec)
        pos = np.dot(fpos, s.get_cell())

        wpoints = [WyckoffPoint(fp, p, o, hprops[h+1])
                   for (fp, p, o, h) in zip(fpos, pos, ops, hess)]

        return wpoints



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of some basic AtomsProperty classes"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from soprano.utils import cart2abc
from soprano.properties import AtomsProperty


class LatticeCart(AtomsProperty):

    """
    LatticeCart

    Property representing the Cartesian form of a structure's lattice

    | Parameters:
    |   shape (tuple): the shape to give to the array

    """

    default_name = 'lattice_cart'
    default_params = {
        'shape': (3, 3)
    }

    @staticmethod
    def extract(s, shape):
        return np.array(s.get_cell()).reshape(shape)


class LatticeABC(AtomsProperty):

    """
    LatticeABC

    Property representing the axis-angles form of a structure's lattice

    | Parameters:
    |   shape (tuple): the shape to give to the array
    |   deg (bool): whether to give the angles in degrees instead of radians

    """

    default_name = 'lattice_abc'
    default_params = {
        'shape': (2, 3),
        'deg': False
    }

    @staticmethod
    def extract(s, shape, deg):
        abc = cart2abc(s.get_cell())
        if deg:
            abc[1, :] *= 180.0/np.pi
        return abc.reshape(shape)


class CalcEnergy(AtomsProperty):

    """
    CalcEnergy

    Property representing the energy calculated by an ASE calulator

    """

    default_name = 'calc_energy'
    default_params = {}

    @staticmethod
    def extract(s):
        try:
            return s.get_potential_energy()
        except:
            return None


class NumAtoms(AtomsProperty):

    """
    NumAtoms

    Property representing the number of atoms in a structure

    """

    default_name = 'num_atoms'
    default_params = {}

    @staticmethod
    def extract(s):
        return s.positions.shape[0]



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility functions for NMR-related properties"""

import re
import numpy as np
import scipy.constants as cnst
from ase.quaternions import Quaternion

# Left here for backwards compatibility
from soprano.data.nmr import (_get_isotope_data, _get_nmr_data, _el_iso)


def _evals_sort(evals, convention='c', return_indices=False):
    """Sort a list of eigenvalue triplets by varios conventions"""
    evals = np.array(evals)
    iso = np.average(evals, axis=1)

    if convention in ('i', 'd'):
        to_sort = evals
    elif convention in ('h', 'n'):
        to_sort = np.abs(evals-iso[:, None])

    sort_i = np.argsort(to_sort,
                        axis=1)
    if convention == 'd':
        sort_i = sort_i[:, ::-1]
    elif convention == 'h':
        sort_i[:, 0], sort_i[:, 1] = sort_i[:, 1], sort_i[:, 0].copy()
    sorted_evals = evals[np.arange(evals.shape[0])[:, None],
                         sort_i]
    if not return_indices:
        return sorted_evals
    else:
        return sorted_evals, sort_i


def _haeb_sort(evals, return_indices=False):
    return _evals_sort(evals, 'h', return_indices)


def _anisotropy(haeb_evals, reduced=False):
    """Calculate anisotropy given eigenvalues sorted with Haeberlen
    convention"""

    f = 2.0/3.0 if reduced else 1.0

    return (haeb_evals[:, 2]-(haeb_evals[:, 0]+haeb_evals[:, 1])/2.0)*f


def _asymmetry(haeb_evals):
    """Calculate asymmetry"""

    aniso = _anisotropy(haeb_evals, reduced=True)
    # Fix the anisotropy zero values
    aniso = np.where(aniso == 0, np.inf, aniso)

    return (haeb_evals[:, 1]-haeb_evals[:, 0])/aniso


def _span(evals):
    """Calculate span"""

    return np.amax(evals, axis=-1)-np.amin(evals, axis=-1)


def _skew(evals):
    """Calculate skew"""

    span = _span(evals)
    span = np.where(span == 0, np.inf, span)
    return 3*(np.median(evals,
                        axis=1) -
              np.average(evals,
                         axis=1))/span


def _evecs_2_quat(evecs):
    """Convert a set of eigenvectors to a Quaternion expressing the
    rotation of the tensor's PAS with respect to the Cartesian axes"""

    # First, guarantee that the eigenvectors express *proper* rotations
    evecs = np.array(evecs)*np.linalg.det(evecs)[:, None, None]

    # Then get the quaternions
    return [Quaternion.from_matrix(evs.T) for evs in evecs]


def _dip_constant(Rij, gi, gj):
    """Dipolar constants for pairs ij, with distances Rij and gyromagnetic
    ratios gi and gj"""

    return - (cnst.mu_0*cnst.hbar*gi*gj / (8*np.pi**2*Rij**3))


def _dip_tensor(d, r, rotation_axis=None):
    """Full dipolar tensor given a constant and a connecting vector"""

    r = np.array(r)
    r /= np.linalg.norm(r)

    if rotation_axis is None:
        D = d*(3*r[:, None]*r[None, :]-np.eye(3))/2.0
    else:
        a = np.array(rotation_axis)
        a /= np.linalg.norm(a)
        vp2 = np.dot(r, a)**2
        D = 0.5*d*(3*vp2-1)*(3*a[:, None]*a[None, :]-np.eye(3))/2.0

    return D


def _J_constant(Kij, gi, gj):
    """J coupling constants for pairs ij, with reduced constant Kij and
    gyromagnetic ratios gi and gj"""

    return cnst.h*gi*gj*Kij/(4*np.pi**2)*1e19



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Providing an interface to selection rules for XRD peaks and various
spacegroups.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import pkgutil

try:
    _xrd_seldata = pkgutil.get_data('soprano',
                                    'data/xrd_sel_rules.json').decode('utf-8')
    xrd_sel_rules = json.loads(_xrd_seldata)
except IOError:
    xrd_sel_rules = None

try:
    _halldata = pkgutil.get_data('soprano',
                                 'data/hall_2_no.json').decode('utf-8')
    hall_2_no = json.loads(_halldata)
except IOError:
    hall_2_no = None


def _ifq(cond, ifT, ifF):
    """ Fortran-style ternary if """
    return ifT if cond else ifF


def _evenq(x):
    """ Returns True if x is even """
    return x % 2 == 0


def _integerq(x):
    """ Returns true if x is integer """
    return int(x) == x


def get_sel_rule_from_international(n, o='all'):
    """ Generate a function object that acts as a selection rule for XRD lines
    for the given symmetry group expressed in international number notation

    | Args:
    |   n (int): International number of the required spacegroup
    |   o (Optional[int]): Sub-option of the required spacegroup

    | Returns:
    |   rule_func (function< list<int> >
    |              => <bool>): a function that can be used to test triples of
    |                          Miller indices h,k,l to verify whether the
    |                          related plane gives rise or not to a peak

    | Raises:
    |   RuntimeError: if the database of XRD selection rules was not properly
    |                 loaded
    |   ValueError: if some of the passed arguments are invalid

    """

    global xrd_sel_rules, _ifq, _evenq, _integerq

    # The JSON keys are strings
    n = str(n)
    o = str(o)

    # First, let's check that we have the correct data,
    # and that the values passed are sensible

    if xrd_sel_rules is None:
        raise RuntimeError("Could not load XRD selection rules")
    if n not in xrd_sel_rules:
        raise ValueError("""Invalid n passed to
                            get_sel_rule_from_international""")

    if o not in xrd_sel_rules[n]:
        if 'all' in xrd_sel_rules[n]:
            o = 'all'
        else:
            if o != 'all':
                raise ValueError("""Invalid o passed to
                                    get_sel_rule_from_international""")
            else:
                raise ValueError("""An option o must be specified for
                                    n = {0}""".format(n))

    xrd_rule_instr = compile(xrd_sel_rules[n][o], '<string>', 'eval')

    return lambda hkl: eval(xrd_rule_instr, {'ifq': _ifq,
                                             'evenq': _evenq,
                                             'integerq': _integerq},
                            {'h': hkl[0],
                             'k': hkl[1],
                             'l': hkl[2]})


def get_sel_rule_from_hall(h):
    """ Generate a function object that acts as a selection rule for XRD lines
    for the given symmetry group expressed in Hall number notation

    | Args:
    |   h (int): Hall number of the required spacegroup

    | Returns:
    |   rule_func (function< list<int> >
    |              => <bool>): a function that can be used to test triples of
    |                          Miller indices h,k,l to verify whether the
    |                          related plane gives rise or not to a peak

    | Raises:
    |   RuntimeError: if the database of XRD selection rules or that of
    |                 Hall numbers was not properly loaded
    |   ValueError: if the passed argument is invalid

    """

    h = str(h)

    if hall_2_no is None:
        raise RuntimeError("""Could not load Hall-to-international conversion
                              table""")
    if h not in hall_2_no:
        raise ValueError("""Invalid h passed to
                            get_sel_rule_from_hall""")

    n = hall_2_no[h]['n']
    o = hall_2_no[h]['o']

    return get_sel_rule_from_international(n, o)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
zcw.py

Contains a class to define the ZCW (Zaremba-Conroy-Wolfsberg) powder averaging scheme.
Implementation taken from:

Edén, M. (2003), Computer simulations in solid‐state NMR. III. Powder averaging. 
Concepts Magn. Reson., 18A: 24-55. doi:10.1002/cmr.a.10065
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from soprano.calculate.powder.powder import PowderScheme


class ZCW(PowderScheme):

    def _calc_engine(self, N):
        # Actually computes the two primary quantities
        # obtained from ZCW: phi and costheta

        # Find the correct g
        zcw_g = [8, 13]

        zcw_c = {'sphere': (1, 2, 1),
                 'hemisphere': (-1, 1, 1),
                 'octant': (-1, 1, 4)}[self.mode]

        # Starting M & N
        zcw_M = 2
        zcw_N = 21
        while zcw_N < N:
            zcw_g.append(zcw_N)
            zcw_M += 1
            zcw_N = zcw_g[-1]+zcw_g[-2]

        # If it's over
        zcw_Nf = 1.0*zcw_N
        zcw_g = zcw_g[-1]

        n = np.arange(0, zcw_N)

        phi = 2.0*np.pi/zcw_c[2]*np.mod(n*zcw_g/zcw_Nf, 1.0)
        ct = zcw_c[0]*(zcw_c[1]*np.mod(n/zcw_Nf, 1.0)-1.0)

        weights = np.ones(len(phi))/len(phi)

        return phi, ct, weights

    def get_orient_angles(self, N):
        """
        Generate and return the ZCW angles (in the form of angles in radians)
        and weights.

        | Args:
        |   N (int): lower bound for the number of orientations generated. The
        |            algorithm will generate at least N orientations.

        | Returns:
        |   angles, weights (np.ndarray): arrays containing respectively the 
        |                                 orientations [theta, phi] and the weights.
        """

        phi, ct, weights = self._calc_engine(N)

        return np.array([np.arccos(ct), phi]).T, weights

    def get_orient_trig(self, N):
        """
        Generate and return the ZCW angles (in the form of trigonometric fuctions)
        and weights.

        | Args:
        |   N (int): lower bound for the number of orientations generated. The
        |            algorithm will generate at least N orientations.

        | Returns:
        |   angles, weights (np.ndarray): arrays containing respectively the 
        |                                 orientations [cos(theta), sin(theta),
        |                                 cos(phi), sin(phi)] and the weights.
        """

        phi, ct, weights = self._calc_engine(N)

        cp = np.cos(phi)
        sp = np.sin(phi)
        st = (1.0-ct**2)**0.5

        return np.array([ct, st, cp, sp]).T, weights

    def get_orient_points(self, N):
        """
        Generate and return the ZCW angles (in the form of points on the surface of
        a sphere) and weights.

        | Args:
        |   N (int): lower bound for the number of orientations generated. The
        |            algorithm will generate at least N orientations.

        | Returns:
        |   angles, weights (np.ndarray): arrays containing respectively the 
        |                                 orientations [x, y, z] and the weights.
        """
        orients, weights = self.get_orient_trig(N)
        ct, st, cp, sp = orients.T
        points = np.array([st*cp, st*sp, ct]).T

        return points, weights
